"""
Copyright 2016-2018 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import absolute_import, division, print_function

import json
import logging
import os
import os.path
import re
import subprocess
import sys
import time
from difflib import unified_diff

import click
from botocore.exceptions import ClientError

from .ef_config import EFConfig
from .ef_service_registry import EFServiceRegistry
from .ef_utils import create_aws_clients, whereami
from .ef_conf_utils import get_account_alias

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler(sys.stderr)
ch.setLevel(logging.INFO)
ch.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
logger.addHandler(ch)

logging.getLogger('botocore').setLevel(logging.CRITICAL)

ret_code = 0
service_registry = None


def diff_string_templates(string_a, string_b):
    """
    Determine the diff of two strings.  Return an empty string if the strings
    are identical, and the diff output string if they are not.
    """
    s1 = string_a.strip().splitlines()
    s2 = string_b.strip().splitlines()
    diffs = unified_diff(s2, s1, fromfile='deployed', tofile='local', lineterm='')
    return '\n'.join(diffs)


def render_local_template(service_name, environment, repo_root, template_file):
    """
    Render a given service's template for a given environment and return it
    """
    cmd = 'cd {} && ef-cf {} {} --devel --verbose'.format(repo_root, template_file, environment)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    if p.returncode != 0:
        stderr = indentify('\n{}'.format(stderr))
        stdout = indentify('\n{}'.format(stdout))
        raise Exception('Service: `{}`, Env: `{}`, Msg: `{}{}`'
                        .format(service_name, environment, stderr, stdout))

    logger.debug('Rendered template for `%s` in `%s`', template_file, environment)

    r = re.match(r".*(^{.*^})$", stdout, re.MULTILINE | re.DOTALL)
    return jsonify(json.loads(r.group(1)))


def fetch_current_cloudformation_template(service_name, environment, cf_client):
    """
    Fetch the currently-deployed template for the given service in the given
    environment and return it.
    """
    stack_name = get_stack_name(environment, service_name)
    logger.debug('Fetching template for `%s`', stack_name)
    result = cf_client.get_template(StackName=stack_name)
    return jsonify(result['TemplateBody'])


def diff_sevice_by_text(service_name, service, environment, cf_client, repo_root):
    """
    Render the local template and compare it to the template that was last
    applied in the target environment.
    """
    global ret_code

    logger.info('Investigating textual diff for `%s`:`%s` in environment `%s`',
                service['type'], service_name, environment)

    try:
        local_template = render_local_template(service_name, environment,
                                               repo_root, service['template_file'])

        current_template = fetch_current_cloudformation_template(
            service_name, environment, cf_client)

    except Exception as e:
        ret_code = 2
        logger.error(e)
        return

    ret = diff_string_templates(local_template, current_template)
    if not ret:
        logger.info('Deployed service `%s` in environment `%s` matches '
                    'the local template.', service_name, environment)
    else:
        ret_code = 1
        logger.error('Service `%s` in environment `%s` differs from '
                     'the local template.',
                     service_name, environment)
        logger.info('Change details:\n        %s', indentify(ret))


def jsonify(dict):
    return json.dumps(dict, indent=2, sort_keys=True)


def indentify(str):
    return str.replace('\n', '\n        ')


def changeset_is_empty(response):
    return (response['Status'] == 'FAILED' and "didn't contain changes" in response['StatusReason'])


def wait_for_changeset_creation(cf_client, changeset_id, changeset_stackid):
    remaining_tries = 30
    while remaining_tries > 0:
        remaining_tries -= 1
        res = cf_client.describe_change_set(ChangeSetName=changeset_id, StackName=changeset_stackid)
        if res['Status'] in ['CREATE_PENDING', 'CREATE_IN_PROGRESS']:
            time.sleep(10)
            continue
        return
    raise Exception('Timed out waiting for changeset to create.')


def generate_changeset(service_name, environment, repo_root, template_file):
    """
    Given a service name and environment, and the details of where the
    template file is, call ef-cf to generate a changeset.  Return the json
    description response that's printed in the ef-cf output.

    Will throw Exception if something goes wrong with the ef-cf call.
    """
    cmd = 'cd {} && ef-cf {} {} --changeset --devel'.format(repo_root, template_file, environment)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    if p.returncode != 0:
        stderr = indentify('\n{}'.format(stderr))
        stdout = indentify('\n{}'.format(stdout))
        raise Exception('Service: `{}`, Env: `{}`, Msg: `{}{}`'
                        .format(service_name, environment, stderr, stdout))

    logger.debug('Created changeset for `%s` in `%s`', template_file, environment)

    r = re.match(r".*^Changeset Info: (.*)$", stdout, re.MULTILINE | re.DOTALL)
    return json.loads(r.group(1))


def get_stack_name(environment, service_name):
    """
    This should be what ef-open templates call "ENV", that is, the short
    name for the environment.
    """
    environment = environment.split('.', 1)[0]
    return '{}-{}'.format(environment, service_name)


def delete_any_existing_changesets(cf_client, service_name, environment):
    stack_name = get_stack_name(environment, service_name)
    logger.debug('Deleting existing changesets for `%s`', stack_name)

    try:
        changesets = cf_client.list_change_sets(StackName=stack_name)
    except ClientError as e:
        logger.debug("Error listing existing changesets: %s", e)
        return

    for changeset in changesets['Summaries']:
        logger.debug('Deleting existing changeset: `%s`', changeset['ChangeSetId'])
        cf_client.delete_change_set(ChangeSetName=changeset['ChangeSetId'], StackName=changeset['StackId'])


def diff_sevice_by_changeset(service_name, service, environment, cf_client, repo_root):
    """
    If an ef-cf call fails, the error will be logged, the retcode set to 2, but
    the function will run to completion and return the list of non-error
    results.
    """
    global ret_code

    logger.info('Investigating changeset for `%s`:`%s` in environment `%s`',
                service['type'], service_name, environment)

    delete_any_existing_changesets(cf_client, service_name, environment)

    try:
        changeset = generate_changeset(service_name, environment,
                                       repo_root, service['template_file'])
    except Exception as e:
        ret_code = 2
        logger.error(e)
        return

    wait_for_changeset_creation(cf_client, changeset['Id'], changeset['StackId'])

    logger.info('Created Changeset ID: `%s`', changeset['Id'])

    desc = cf_client.describe_change_set(
        ChangeSetName=changeset['Id'], StackName=changeset['StackId'])

    cf_client.delete_change_set(
        ChangeSetName=changeset['Id'], StackName=changeset['StackId'])

    if changeset_is_empty(desc):
        logger.info('Deployed service `%s` in environment `%s` matches '
                    'the local template.', service_name, environment)
    else:
        ret_code = 1
        logger.error('Service `%s` in environment `%s` differs from '
                     'the local template.',
                     service_name, environment)
        details = jsonify(desc['Changes'])
        logger.info('Change details:\n        %s', indentify(details))


def get_cloudformation_client(service_name, environment_name):
    """
    Given a service name and an environment name, return a boto CloudFormation
    client object.
    """
    region = service_registry.service_region(service_name)

    if whereami() == 'ec2':
        profile = None
    else:
        profile = get_account_alias(environment_name)

    clients = create_aws_clients(region, profile, 'cloudformation')
    return clients['cloudformation']


def generate_test_environment_name(env_name):
    """
    Some environments, like proto, need a numerical index, while others, like
    prod do not.  Given an environment name from the service registry, return a
    valid environment name in which to generate a template.
    """
    if env_name in ['alpha', 'proto']:
        return '{}0'.format(env_name)
    return env_name


def get_env_categories(envs):
    """
    Given a list of environments, return an associated list of environment
    base names where, for example, 'proto0' becomes 'proto', but 'staging'
    remains 'staging'.
    """
    return [re.match(r'^(.*?)\d*$', name).group(1) for name in envs]


def evaluate_service_changes(services, envs, repo_root, func):
    """
    Given a dict of services, and a list of environments, apply the diff
    function to evaluate the differences between the target environments
    and the rendered templates.

    Sub-services (names with '.' in them) are skipped.
    """
    for service_name, service in services.iteritems():

        for env_category in service['environments']:
            if env_category not in get_env_categories(envs):
                logger.debug('Skipping not-included environment `%s` for service `%s`',
                             env_category, service_name)
                continue

            environment = generate_test_environment_name(env_category)

            cf_client = get_cloudformation_client(service_name, environment)

            func(service_name, service, environment, cf_client, repo_root)


def test_for_unused_template_files(template_files, services):
    """
    Loop over the template files, and print a warning for any that aren't
    used by a service in the service registry
    (the annoying sub-function is to deal with Python's lack of a labeled break)
    """
    def print_unused_template_warning(file, services):
        for service_name, service in services.iteritems():
            if service['template_file'] == file:
                return
        logger.warning("Template file has no service registry entry: `%s`", file)

    for name, file in template_files.iteritems():
        print_unused_template_warning(file, services)


def get_matching_service_template_file(service_name, template_files):
    """
    Return the template file that goes with the given service name, or return
    None if there's no match.  Subservices return the parent service's file.
    """
    # If this is a subservice, use the parent service's template
    service_name = service_name.split('.')[0]
    if service_name in template_files:
        return template_files[service_name]
    return None


def get_dict_registry_services(registry, template_files, warn_missing_files=True):
    """
    Return a dict mapping service name to a dict containing the service's
    type ('fixtures', 'platform_services', 'application_services', 'internal_services'),
    the template file's absolute path, and a list of environments to which the
    service is intended to deploy.

    Service names that appear twice in the output list will emit a warning and
    ignore the latter records.

    Services which have no template file will not appear in the returned dict.
    If the `warn_missing_files` boolean is True these files will emit a warning.
    """
    with open(registry) as fr:
        parsed_registry = json.load(fr)

    services = {}
    for type, type_services in parsed_registry.iteritems():
        for name, service in type_services.iteritems():
            if name in services:
                logger.warning("Template name appears twice, ignoring later items: `%s`", name)
                continue

            template_file = get_matching_service_template_file(name, template_files)
            if not template_file:
                if warn_missing_files:
                    logger.warning("No template file for `%s` (%s) `%s`", type, service['type'], name)
                continue

            services[name] = {
                'type': type,
                'template_file': template_file,
                'environments': service['environments']
            }

    return services


def scan_dir_for_template_files(search_dir):
    """
    Return a map of "likely service/template name" to "template file".
    This includes all the template files in fixtures and in services.
    """
    template_files = {}
    cf_dir = os.path.join(search_dir, 'cloudformation')
    for type in os.listdir(cf_dir):
        template_dir = os.path.join(cf_dir, type, 'templates')
        for x in os.listdir(template_dir):
            name = os.path.splitext(x)[0]
            template_files[name] = os.path.join(template_dir, x)
    return template_files


@click.command()
@click.option('--repo_root',
              default="./",
              required=False,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True, resolve_path=True),
              help="The root directory of the template repository git clone.")
@click.option('--sr',
              default=None,
              required=False,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              readable=True, resolve_path=True),
              help="optional /path/to/service_registry_file.json")
@click.option('--env', '-e',
              multiple=True,
              required=True,
              type=click.Choice(EFConfig.ENV_LIST),
              help="An environment to evaluate. Can be set several times.")
@click.option('--template_file', '-t',
              multiple=True,
              required=False,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              readable=True, resolve_path=True),
              help="A specific template to process.  Can be passed multiple "
                   "times.  If excluded, all templates will be run.")
@click.option('--raw_text', '-r',
              is_flag=True,
              help="Instead of checking changesets, compare the current text of "
                   "the template(s) with the last pushed version.")
@click.version_option()
def main(repo_root, sr, env, template_file, raw_text):
    """Generate a diff based on cloudformation template specified and what's currently deployed."""
    global service_registry
    service_registry = EFServiceRegistry(sr)

    template_files = scan_dir_for_template_files(repo_root)

    # If the list of whitelisted templates was passed, we're in "only test
    # those templates" mode.  Otherwise we'll just do every template.
    if template_file:
        template_files = {name: file
                          for name, file in template_files.iteritems()
                          if file in template_file}

    services = get_dict_registry_services(service_registry.filespec,
                                          template_files,
                                          warn_missing_files=(not template_file))

    test_for_unused_template_files(template_files, services)

    if raw_text:
        evaluate_service_changes(services, env, repo_root, diff_sevice_by_text)
    else:
        evaluate_service_changes(services, env, repo_root, diff_sevice_by_changeset)

    exit(ret_code)



#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function
from collections import OrderedDict
import argparse
import json
import os
import string
import sys

from ef_config import EFConfig
from ef_context import EFContext
import ef_utils


class EFPWContext(EFContext):
  def __init__(self):
    super(EFPWContext, self).__init__()
    self._decrypt = None
    self._re_encrypt = None
    self._length = 32
    self._plaintext = None
    self._secret_file = None
    self._match = None
    self._special_character_override = None
    self._forbidden_characters = ['§', '±', '\\', '>', '<', ':', '\'', '"', '`']

  @property
  def decrypt(self):
    """String value if the tool should decrypt rather than generating a new encrypted secret"""
    return self._decrypt

  @decrypt.setter
  def decrypt(self, value):
    if type(value) is not str:
      raise TypeError("decrypt value must be str")
    self._decrypt = value

  @property
  def re_encrypt(self):
    """String value if the tool should re_encrypt rather than generating a new encrypted secret"""
    return self._re_encrypt

  @re_encrypt.setter
  def re_encrypt(self, value):
    if type(value) is not str:
      raise TypeError("re_encrypt value must be str")
    self._re_encrypt = value

  @property
  def length(self):
    """Integer length of the secret to be generated"""
    return self._length

  @length.setter
  def length(self, value):
    if type(value) is not int and not value.isdigit():
      raise ValueError("length value must be int")
    elif int(value) < 10:
      raise ValueError("length value must be >= 10")
    else:
      self._length = int(value)

  @property
  def plaintext(self):
    """String value of the user-provided secret to be encrypted"""
    return self._plaintext

  @plaintext.setter
  def plaintext(self, value):
    if type(value) is not str:
      raise TypeError("plaintext value must be str")
    if sys.getsizeof(value) > 4096:
      raise ValueError("plaintext value may not be larger than 4kb")
    self._plaintext = value

  @property
  def secret_file(self):
    """String value of the user-provided secret file to be encrypted"""
    return self._secret_file

  @secret_file.setter
  def secret_file(self, value):
    if type(value) is not str:
      raise TypeError("secret_file value must be str")
    self._secret_file = value

  @property
  def match(self):
    """String value to match against in the secret file"""
    return self._match

  @match.setter
  def match(self, value):
    if type(value) is not str:
      raise TypeError("match value must be str")
    self._match = value

  @property
  def special_character_override(self):
    """Special characters to dissmiss when entering a plain-text to be encrypted"""
    return self._special_character_override

  @special_character_override.setter
  def special_character_override(self, value):
    if type(value) is not str:
      raise TypeError("special_character_override value must be string")
    self._special_character_override = value

  @property
  def forbidden_characters(self):
    return self._forbidden_characters

def format_secret(secret):
  """
  Format secret to compatible decrypt string
  Args:
    secret (string): KMS secret hash
  Returns:
    formatted ef resolvable KMS decrypt string
  Raises:
    None
  """
  return "{{aws:kms:decrypt,%s}}" % secret

def generate_secret(length=32):
  """
  Generate a random secret consisting of mixed-case letters and numbers
  Args:
      length (int): Length of the generated password
  Returns:
      a randomly generated secret string
  Raises:
      None
  """
  alphabet = string.ascii_letters + string.digits
  random_bytes = os.urandom(length)
  indices = [int(len(alphabet) * (ord(byte) / 256.0)) for byte in random_bytes]
  return "".join([alphabet[index] for index in indices])

def generate_secret_file(file_path, pattern, service, environment, clients):
  """
  Generate a parameter files with it's secrets encrypted in KMS
  Args:
      file_path (string): Path to the parameter file to be encrypted
      pattern (string): Pattern to do fuzzy string matching
      service (string): Service to use KMS key to encrypt file
      environment (string): Environment to encrypt values
      clients (dict): KMS AWS client that has been instantiated
  Returns:
      None
  Raises:
    IOError: If the file does not exist
  """
  changed = False
  with open(file_path) as json_file:
    data = json.load(json_file, object_pairs_hook=OrderedDict)
    try:
      for key, value in data["params"][environment].items():
        if pattern in key:
          if "aws:kms:decrypt" in value:
            print("Found match, key {} but value is encrypted already; skipping...".format(key))
          else:
            print("Found match, encrypting key {}".format(key))
            encrypted_password = ef_utils.kms_encrypt(clients['kms'], service, environment, value)
            data["params"][environment][key] = format_secret(encrypted_password)
            changed = True
    except KeyError:
      ef_utils.fail("Error env: {} does not exist in parameters file".format(environment))

  if changed:
    with open(file_path, "w") as encrypted_file:
      json.dump(data, encrypted_file, indent=2, separators=(',', ': '))
      # Writing new line here so it conforms to WG14 N1256 5.1.1.1 (so github doesn't complain)
      encrypted_file.write("\n")

def handle_args_and_set_context(args):
  """
  Args:
      args: the command line args, probably passed from main() as sys.argv[1:]
  Returns:
      a populated EFPWContext object
  Raises:
      RuntimeError: if branch isn't as spec'd in ef_config.EF_REPO_BRANCH
      ValueError: if a parameter is invalid
  """
  parser = argparse.ArgumentParser(description="Encrypt/decrypt template secrets.")
  parser.add_argument("service", help="name of service password is being generated for")
  parser.add_argument("env", help=", ".join(EFConfig.ENV_LIST))
  parser.add_argument("--special_character_override", help="comma-separated list of characters \
    to override in quotes - only use with plaintext argument to allow a special charater set.\
    Pick a subset from: §,±,<,>,',\",`,\\,: which are denied by default" , default="")
  group = parser.add_mutually_exclusive_group()
  group.add_argument("--decrypt", help="encrypted string to be decrypted", default="")
  group.add_argument("--re-encrypt", help="encrypted string to be re encrypted for a new service", default="")
  group.add_argument("--plaintext", help="secret to be encrypted rather than a randomly generated one", default="")
  group.add_argument("--secret_file", help="json file containing secrets to be encrypted", default="")
  parser.add_argument("--match", help="used in conjunction with --secret_file to match against keys to be encrypted", default="")
  parser.add_argument("--length", help="length of generated password (default 32)", default=32)
  parser.add_argument("--lbv1_escapes", help="if secret has escapes, encode them for lbv1 consumption", action="store_true", default=False)
  parsed_args = vars(parser.parse_args(args))
  context = EFPWContext()

  try:
    context.env = parsed_args["env"]
  except ValueError as e:
    ef_utils.fail("Error in env: {}".format(e))

  context.service = parsed_args["service"]
  context.decrypt = parsed_args["decrypt"]
  context.re_encrypt = parsed_args["re_encrypt"]
  context.length = parsed_args["length"]
  context.plaintext = parsed_args["plaintext"]
  if not parsed_args["lbv1_escapes"]:
    # unescape any escapes that the shell might've added; e.g: \\n becomes \n
    context.plaintext = parsed_args["plaintext"].decode("string_escape")
  context.secret_file = parsed_args["secret_file"]
  context.match = parsed_args["match"]
  if context.match or context.secret_file:
    if not context.match or not context.secret_file:
      raise ValueError("Must have both --match and --secret_file flag")
  context.special_character_override = parsed_args["special_character_override"]
  #special_characters_override only applies when plaintext is present
  if (context.plaintext is None or context.plaintext == '') and context.special_character_override:
    parser.error("special_character_override parameter can only be specified when plaintext parameter is specified")
  if context.special_character_override:
    for special_char in context.special_character_override.split(","):
      special_char = special_char.strip()
      if special_char not in context.forbidden_characters:
        parser.error("{} special character is part of forbidden characters list: {}".format(special_char, ', '.join(context.forbidden_characters)))
  return context


def main():
  context = handle_args_and_set_context(sys.argv[1:])
  profile = None if context.whereami == "ec2" else context.account_alias

  try:
    clients = ef_utils.create_aws_clients(EFConfig.DEFAULT_REGION, profile, "kms")
  except RuntimeError as error:
    ef_utils.fail(
      "Exception creating clients in region {} with profile {}".format(EFConfig.DEFAULT_REGION, profile),
      error
    )

  if context.secret_file:
    generate_secret_file(context.secret_file, context.match, context.service, context.env, clients)
    return

  if context.decrypt:
    decrypted = ef_utils.kms_decrypt(kms_client=clients['kms'], secret=context.decrypt)
    key_aliases = ef_utils.kms_key_alias(clients['kms'], decrypted.key_id)
    print("Decrypted Secret: {}; Key: {}".format(decrypted.plaintext, ', '.join(key_aliases)))
    return

  if context.re_encrypt:
    encrypted_password = ef_utils.kms_re_encrypt(
      kms_client=clients['kms'],
      service=context.service,
      env=context.env,
      secret=context.re_encrypt)
    print(format_secret(encrypted_password))
    return

  if context.plaintext:
    password = context.plaintext
    for forbidden_char in context.forbidden_characters:
      if forbidden_char in password and forbidden_char not in context.special_character_override:
        raise ValueError(
            "Character {} is forbidden. Use the argument --special_character_override to bypass it.".format(forbidden_char))
  else:
    password = generate_secret(context.length)
    print("Generated Secret: {}".format(password))
  encrypted_password = ef_utils.kms_encrypt(clients['kms'], context.service, context.env, password)
  print(format_secret(encrypted_password))
  return

if __name__ == "__main__":
  main()



"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import datetime
import re

from botocore.exceptions import ClientError

import ef_utils

class EFAwsResolver(object):
  """
  For keys to look up, we use partial ARN syntax to identify system and information sought:
  http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#arn-syntax-ec2
  Not all possible lookups are supported

  Expects these clients to be pre-made and passed in:
    cloudfront, cloudformation, ec2, iam, kms, lambda, route53, waf

  Example:
    the ARN of a security group ID is:
      arn:aws:ec2:us-west-2:1234567890:security-group/security-group-id
    following this pattern:
      arn:aws:<service>:<region>:<account-id>:<key>
    we trim it to:
      <service>:<key>
      ec2:security-group/security-group-id
    which becomes a template token with 'aws:' prepended to identify the lookup provider,
    and the search string appended:
      {{aws:ec2:security-group/security-group-id,proto2-ess-elb}}
    which then is passed to the lookup function here as:
      EFAwsResolver.lookup("ec2:security-group/security-group-id","proto2-ess-elb")
  """

  # dictionary of boto3 clients: {"ec2":ec2_client, ... } made with ef_utils.create_aws_clients
  __CLIENTS = {}

  def _elbv2_load_balancer(self, lookup):
    """
    Args:
      lookup: the friendly name of the V2 elb to look up
    Returns:
      A dict with the load balancer description
    Raises:
      botocore.exceptions.ClientError: no such load-balancer
    """
    client = EFAwsResolver.__CLIENTS['elbv2']
    elbs = client.describe_load_balancers(Names=[lookup])
    # getting the first one, since we requested only one lb
    elb = elbs['LoadBalancers'][0]
    return elb

  def acm_certificate_arn(self, lookup, default=None):
    """
    Args:
      lookup: region/domain on the certificate to be looked up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      ARN of a certificate with status "Issued" for the region/domain, if found, or default/None if no match
      If more than one "Issued" certificate matches the region/domain:
        - if any matching cert was issued by Amazon, returns ARN of certificate with most recent IssuedAt timestamp
        - if no certs were issued by Amazon, returns ARN of an arbitrary matching certificate
        - certificates issued by Amazon take precedence over certificates not issued by Amazon
    """
    # @todo: Only searches the first 100 certificates in the account

    try:
      # This a region-specific client, so we'll make a new client in the right place using existing SESSION
      region_name, domain_name = lookup.split("/")
      acm_client = EFAwsResolver.__CLIENTS["SESSION"].client(service_name="acm", region_name=region_name)
      response = acm_client.list_certificates(
        CertificateStatuses=['ISSUED'],
        MaxItems=100
      )
    except Exception:
      return default
    # No certificates
    if len(response["CertificateSummaryList"]) < 1:
      return default
    # One or more certificates - find cert with latest IssuedAt date or an arbitrary cert if none are dated
    best_match_cert = None
    for cert_handle in response["CertificateSummaryList"]:
      if cert_handle["DomainName"] == domain_name:
        cert = acm_client.describe_certificate(CertificateArn=cert_handle["CertificateArn"])["Certificate"]
        # Patch up cert if there is no IssuedAt (i.e. cert was not issued by Amazon)
        if "IssuedAt" not in cert:
          cert[u"IssuedAt"] = datetime.datetime(1970, 1, 1, 0, 0)
        if best_match_cert is None:
          best_match_cert = cert
        elif cert["IssuedAt"] > best_match_cert["IssuedAt"]:
          best_match_cert = cert
    if best_match_cert is not None:
      return best_match_cert["CertificateArn"]
    return default

  def ec2_elasticip_elasticip_id(self, lookup, default=None):
    """
    Args:
      lookup: the CloudFormation resource name of the Elastic IP ID to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The ID of the first Elastic IP found with a description matching 'lookup' or default/None if no match found
    """

    public_ip = self.ec2_elasticip_elasticip_ipaddress(lookup)
    if public_ip is None:
      return default
    try:
      eips = EFAwsResolver.__CLIENTS["ec2"].describe_addresses(
        PublicIps=[public_ip]
      )
    # Public IP not found
    except ClientError:
      return default
    eip_id = eips["Addresses"][0]["AllocationId"]
    return eip_id

  def ec2_elasticip_elasticip_ipaddress(self, lookup, default=None):
    """
    Args:
      lookup: the CloudFormation resource name of the Elastic IP address to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The IP address of the first Elastic IP found with a description matching 'lookup' or default/None if no match
    """
    # Extract environment from resource ID to build stack name
    m = re.search('ElasticIp([A-Z]?[a-z]+[0-9]?)\w+', lookup)
    # The lookup string was not a valid ElasticIp resource label
    if m is None:
      return default
    env = m.group(1)
    stackname = "{}-elasticip".format(env.lower())
    # Convert env substring to title in case {{ENV}} substitution is being used
    lookup = lookup.replace(env, env.title())
    # Look up the EIP resource in the stack to get the IP address assigned to the EIP
    try:
      eip_stack = EFAwsResolver.__CLIENTS["cloudformation"].describe_stack_resources(
        StackName=stackname,
        LogicalResourceId=lookup
      )
    except ClientError:
      return default
    stack_resources = eip_stack["StackResources"]
    # Resource does not exist in stack
    if len(stack_resources) < 1:
      return default
    eip_publicip = stack_resources[0]["PhysicalResourceId"]
    return eip_publicip

  def ec2_eni_eni_id(self, lookup, default=None):
    """
    Args:
      lookup: the description of the Elastic Network Interface (ENI) to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The ID of the first ENI found with a description matching 'lookup' or default/None if no match found
    """
    enis = EFAwsResolver.__CLIENTS["ec2"].describe_network_interfaces(Filters=[{
      'Name': 'description',
      'Values': [lookup]
    }])
    if len(enis.get("NetworkInterfaces")) > 0:
      return enis["NetworkInterfaces"][0]["NetworkInterfaceId"]
    else:
      return default

  def ec2_network_network_acl_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the network ACL we are looking up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the network ACL, or None if no match found
    """
    network_acl_id = EFAwsResolver.__CLIENTS["ec2"].describe_network_acls(Filters=[{
      'Name': 'tag:Name',
      'Values': [lookup]
    }])
    if len(network_acl_id["NetworkAcls"]) > 0:
      return network_acl_id["NetworkAcls"][0]["NetworkAclId"]
    else:
      return default

  def ec2_security_group_security_group_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of a security group to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      Security group ID if target found or default/None if no match
    """
    try:
      response = EFAwsResolver.__CLIENTS["ec2"].describe_security_groups(Filters=[{
        'Name':'group-name', 'Values':[lookup]
      }])
    except:
      return default
    if len(response["SecurityGroups"]) > 0:
      return response["SecurityGroups"][0]["GroupId"]
    else:
      return default

  def ec2_subnet_subnet_id(self, lookup, default=None):
    """
    Return:
      the ID of a single subnet or default/None if no match
    Args:
      lookup: the friendly name of the subnet to look up (subnet-<env>-a or subnet-<env>-b)
      default: the optional value to return if lookup failed; returns None if not set
    """
    subnets = EFAwsResolver.__CLIENTS["ec2"].describe_subnets(Filters=[{
      'Name': 'tag:Name',
      'Values': [lookup]
    }])
    if len(subnets["Subnets"]) > 0:
      return subnets["Subnets"][0]["SubnetId"]
    else:
      return default

  def ec2_subnet_subnet_cidr(self, lookup, default=None):
    """
    Return:
      the ID of a single subnet or default/None if no match
    Args:
      lookup: the friendly name of the subnet to look up (subnet-<env>-a or subnet-<env>-b)
      default: the optional value to return if lookup failed; returns None if not set
    """
    subnets = EFAwsResolver.__CLIENTS["ec2"].describe_subnets(Filters=[{
      'Name': 'tag:Name',
      'Values': [lookup]
    }])
    if len(subnets["Subnets"]) > 0:
      return subnets["Subnets"][0]["CidrBlock"]
    else:
      return default

  def ec2_vpc_availabilityzones(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of a VPC to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      A comma-separated list of availability zones in use in the named VPC or default/None if no match
    """
    vpc_id = self.ec2_vpc_vpc_id(lookup)
    if vpc_id is None:
      return default
    subnets = EFAwsResolver.__CLIENTS["ec2"].describe_subnets(Filters=[{
      'Name': 'vpc-id',
      'Values': [vpc_id]
    }])
    if len(subnets["Subnets"]) > 0:
      # Strip the metadata section (subnets["Subnets"])
      az_list = [s["AvailabilityZone"] for s in subnets["Subnets"]]
      # Add internal ", " only. This is called literally from: "{{aws...}}" - CF template needs the outer quotes
      return "\", \"".join(az_list)
    else:
      return default

  def ec2_vpc_subnets(self, lookup, default=None):
    """
    Args:
      lookup - the friendly name of the VPC whose subnets we want
    Returns:
      A comma-separated list of all subnets in use in the named VPC or default/None if no match found
    """
    vpc_id = self.ec2_vpc_vpc_id(lookup)
    if vpc_id is None:
      return default
    subnets = EFAwsResolver.__CLIENTS["ec2"].describe_subnets(Filters=[{
      'Name': 'vpc-id',
      'Values': [vpc_id]
    }])
    if len(subnets["Subnets"]) > 0:
      # Strip the metadata section (subnets["Subnets"])
      subnet_list = [s["SubnetId"] for s in subnets["Subnets"]]
      # Add internal ", " only. This is called literally from: "{{aws...}}" - reuses the outer quotes
      return "\", \"".join(subnet_list)
    else:
      return default

  def ec2_vpc_cidrblock(self, lookup, default=None):
    """
    Args:
      lookup - the friendly name of the VPC whose CIDR block we want
    Returns:
      The CIDR block of the named VPC, or default/None if no match found
    """
    vpcs = EFAwsResolver.__CLIENTS["ec2"].describe_vpcs(Filters=[{
      'Name': 'tag:Name',
      'Values': [lookup]
    }])
    if len(vpcs.get("Vpcs")) > 0:
      return vpcs["Vpcs"][0]["CidrBlock"]
    else:
      return default

  def ec2_vpc_vpc_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the VPC to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The ID of the first VPC found with a label matching 'lookup' or default/None if no match found
    """
    vpcs = EFAwsResolver.__CLIENTS["ec2"].describe_vpcs(Filters=[{
      'Name': 'tag:Name',
      'Values': [lookup]
    }])
    if len(vpcs.get("Vpcs")) > 0:
      return vpcs["Vpcs"][0]["VpcId"]
    else:
      return default

  def ec2_vpc_endpoint_id(self, lookup, default=None):
    """
    Args:
      lookup: the name of the VPC endpoint to look up (in tags)
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The ID of the VPC Endpoint found with a label matching 'lookup' or default/None if no match found
    """
    vpc_endpoints = EFAwsResolver.__CLIENTS["ec2"].describe_vpc_endpoints(Filters=[{
      "Name": "tag:Name",
      "Values": [lookup]
    }])
    if len(vpc_endpoints.get("VpcEndpoints")) > 0:
      return vpc_endpoints["VpcEndpoints"][0]["VpcEndpointId"]
    else:
      return default

  def ec2_vpc_endpoint_id_by_vpc_service(self, lookup, default=None):
    """
    Args:
      lookup: a forward-slash-delimited string of [vpc-name, service-name] "vpc-name/service-name"
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The ID of the OLDEST VPC endpoint found in the given VPC for the given service
    """
    vpc_name, service_name = lookup.split("/")
    vpc_id = self.ec2_vpc_vpc_id(vpc_name)
    if vpc_id is None:
      return default

    vpc_endpoints = EFAwsResolver.__CLIENTS["ec2"].describe_vpc_endpoints(Filters=[
      {"Name": "vpc-id", "Values":[vpc_id]},
      {"Name": "service-name", "Values":[service_name]}
    ])
    if len(vpc_endpoints.get("VpcEndpoints")) < 1:
      return default

    oldest = None
    for vpce in vpc_endpoints.get("VpcEndpoints"):
      if oldest is None or vpce["CreationTimestamp"] < oldest["CreationTimestamp"]:
        oldest = vpce
    return oldest["VpcEndpointId"]

  def ec2_vpc_endpoint_dns_name(self, lookup, default=None):
    """
    Args:
      lookup: the name of the VPC endpoint to look up (in tags)
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The first DNS Name of the VPC Endpoint found with a label matching 'lookup' or default/None if no match found
    """
    vpc_endpoints = EFAwsResolver.__CLIENTS["ec2"].describe_vpc_endpoints(Filters=[{
      "Name": "tag:Name",
      "Values": [lookup]
    }])
    if len(vpc_endpoints.get("VpcEndpoints")) > 0:
      if len(vpc_endpoints["VpcEndpoints"][0]["DnsEntries"]) < 1:
        return default
      return vpc_endpoints["VpcEndpoints"][0]["DnsEntries"][0]["DnsName"]
    else:
      return default

  def ec2_vpc_endpoint_dns_name_by_vpc_service(self, lookup, default=None):
    """
    Args:
      lookup: a forward-slash-delimited string of [vpc-name, service-name] "vpc-name/service-name"
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The first DNS Name of the OLDEST VPC endpoint found in the given VPC for the given service
    """
    vpc_name, service_name = lookup.split("/")
    vpc_id = self.ec2_vpc_vpc_id(vpc_name)
    if vpc_id is None:
      return default

    vpc_endpoints = EFAwsResolver.__CLIENTS["ec2"].describe_vpc_endpoints(Filters=[
      {"Name": "vpc-id", "Values":[vpc_id]},
      {"Name": "service-name", "Values":[service_name]}
    ])
    if len(vpc_endpoints.get("VpcEndpoints")) < 1:
      return default

    oldest = None
    for vpce in vpc_endpoints.get("VpcEndpoints"):
      if oldest is None or vpce["CreationTimestamp"] < oldest["CreationTimestamp"]:
        oldest = vpce

    if len(oldest["DnsEntries"]) < 1:
      return default
    return oldest["DnsEntries"][0]["DnsName"]

  def ec2_vpc_vpn_gateway_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the VPN Gateway ID to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The ID of the VPN Gateway found with a label matching 'lookup' or default/None if no match found
    """
    vpn_gateways = EFAwsResolver.__CLIENTS["ec2"].describe_vpn_gateways(Filters=[{
      "Name": "tag:Name",
      "Values": [lookup]
    }])
    if len(vpn_gateways) > 0:
      return vpn_gateways["VpnGateways"][0]["VpnGatewayId"]
    else:
      return default

  def elbv2_load_balancer_hosted_zone(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the V2 elb to look up
      default: value to return in case of no match
    Returns:
      The hosted zone ID of the ELB found with a name matching 'lookup'.
    """
    try:
      elb = self._elbv2_load_balancer(lookup)
      return elb['CanonicalHostedZoneId']
    except ClientError:
      return default

  def elbv2_load_balancer_dns_name(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the V2 elb to look up
      default: value to return in case of no match
    Returns:
      The hosted zone ID of the ELB found with a name matching 'lookup'.
    """
    try:
      elb = self._elbv2_load_balancer(lookup)
      return elb['DNSName']
    except ClientError:
      return default

  def elbv2_load_balancer_arn_suffix(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the v2 elb to look up
      default: value to return in case of no match
    Returns:
      The shorthand fragment of the ALB's ARN, of the form `app/*/*`
    """
    try:
      elb = self._elbv2_load_balancer(lookup)
      m = re.search(r'.+?(app\/[^\/]+\/[^\/]+)$', elb['LoadBalancerArn'])
      return m.group(1)
    except ClientError:
      return default

  def elbv2_target_group_arn_suffix(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the v2 elb target group
      default: value to return in case of no match
    Returns:
      The shorthand fragment of the target group's ARN, of the form
      `targetgroup/*/*`
    """
    try:
      client = EFAwsResolver.__CLIENTS['elbv2']
      elbs = client.describe_target_groups(Names=[lookup])
      elb = elbs['TargetGroups'][0]
      m = re.search(r'.+?(targetgroup\/[^\/]+\/[^\/]+)$', elb['TargetGroupArn'])
      return m.group(1)
    except ClientError:
      return default

  def waf_ip_set_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of an IP set
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the IP set whose name matches 'lookup' or default/None if no match found
    """
    # list_rules returns at most 100 rules per request
    list_limit = 100
    list_ip_sets = EFAwsResolver.__CLIENTS["waf"].list_ip_sets
    ip_sets = list_ip_sets(Limit=list_limit)
    while True:
      for rule in ip_sets["IPSets"]:
        if rule["Name"] == lookup:
          return rule["IPSetId"]
      if "NextMarker" in ip_sets:
        ip_sets = list_ip_sets(Limit=list_limit, NextMarker=ip_sets["NextMarker"])
      else:
        return default

  def waf_rule_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of a WAF rule
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the WAF rule whose name matches 'lookup' or default/None if no match found
    """
    # list_rules returns at most 100 rules per request
    list_limit = 100
    rules = EFAwsResolver.__CLIENTS["waf"].list_rules(Limit=list_limit)
    while True:
      for rule in rules["Rules"]:
        if rule["Name"] == lookup:
          return rule["RuleId"]
      if "NextMarker" in rules:
        rules = EFAwsResolver.__CLIENTS["waf"].list_rules(Limit=list_limit, NextMarker=rules["NextMarker"])
      else:
        return default

  def waf_web_acl_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of a Web ACL
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the WAF Web ACL whose name matches rule_name or default/None if no match found
    """
    # list_rules returns at most 100 rules per request
    list_limit = 100
    acls = EFAwsResolver.__CLIENTS["waf"].list_web_acls(Limit=list_limit)
    while True:
      for acl in acls["WebACLs"]:
        if acl["Name"] == lookup:
          return acl["WebACLId"]
      if "NextMarker" in acls:
        acls = EFAwsResolver.__CLIENTS["waf"].list_web_acls(Limit=list_limit, NextMarker=acls["NextMarker"])
      else:
        return default

  def route53_public_hosted_zone_id(self, lookup, default=None):
    """
    Args:
      lookup: The zone name to look up. Must end with "."
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the public hosted zone for the 'lookup' domain, or default/None if no match found
    """
    list_limit = "100"
    # enforce terminal '.' in name, otherwise we could get a partial match of the incorrect zones
    if lookup[-1] != '.':
      return default
    hosted_zones = EFAwsResolver.__CLIENTS["route53"].list_hosted_zones_by_name(DNSName=lookup, MaxItems=list_limit)
    # Return if the account has no HostedZones
    if "HostedZones" not in hosted_zones:
      return default
    while True:
      for hosted_zone in hosted_zones["HostedZones"]:
        if lookup == hosted_zone["Name"] and not hosted_zone["Config"]["PrivateZone"]:
          return hosted_zone["Id"].split("/")[2]
      if hosted_zones["IsTruncated"]:
        hosted_zones = EFAwsResolver.__CLIENTS["route53"].list_hosted_zones_by_name(
          DNSName=hosted_zones["NextDNSName"], HostedZoneId=hosted_zones["NextHostedZoneId"], MaxItems=list_limit)
      else:
        return default

  def route53_private_hosted_zone_id(self, lookup, default=None):
    """
    Args:
      lookup: The zone name to look up. Must end with "."
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the private hosted zone for the 'lookup' domain, or default/None if no match found
    """
    list_limit = "100"
    # enforce terminal '.' in name, otherwise we could get a partial match of the incorrect zones
    if lookup[-1] != '.':
      return default
    hosted_zones = EFAwsResolver.__CLIENTS["route53"].list_hosted_zones_by_name(DNSName=lookup, MaxItems=list_limit)
    # Return if the account has no HostedZones
    if "HostedZones" not in hosted_zones:
      return default
    while True:
      for hosted_zone in hosted_zones["HostedZones"]:
        if lookup == hosted_zone["Name"] and hosted_zone["Config"]["PrivateZone"]:
          return hosted_zone["Id"].split("/")[2]
      if hosted_zones["IsTruncated"]:
        hosted_zones = EFAwsResolver.__CLIENTS["route53"].list_hosted_zones_by_name(
          DNSName=hosted_zones["NextDNSName"], HostedZoneId=hosted_zones["NextHostedZoneId"], MaxItems=list_limit)
      else:
        return default

  def ec2_route_table_main_route_table_id(self, lookup, default=None):
    """
    Args:
      lookup: the friendly name of the VPC whose main route table we are looking up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the main route table of the named VPC, or default if no match/multiple matches found
    """
    vpc_id = self.ec2_vpc_vpc_id(lookup)
    if vpc_id is None:
      return default
    route_table = EFAwsResolver.__CLIENTS["ec2"].describe_route_tables(Filters=[
      {'Name': 'vpc-id', 'Values': [vpc_id]},
      {'Name': 'association.main', 'Values': ['true']}
    ])
    if len(route_table["RouteTables"]) is not 1:
      return default
    return route_table["RouteTables"][0]["RouteTableId"]

  def ec2_route_table_tagged_route_table_id(self, lookup, default=None):
    """
    Args:
      lookup: the tagged route table name, should be unique
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the route table, or default if no match/multiple matches found
    """
    route_table = EFAwsResolver.__CLIENTS["ec2"].describe_route_tables(Filters=[
      {'Name': 'tag-key', 'Values': ['Name']},
      {'Name': 'tag-value', 'Values': [lookup]}
    ])
    if len(route_table["RouteTables"]) is not 1:
      return default
    return route_table["RouteTables"][0]["RouteTableId"]

  def cloudfront_domain_name(self, lookup, default=None):
    """
    Args:
      lookup: any CNAME on the Cloudfront distribution
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The domain name (FQDN) of the Cloudfront distrinbution, or default/None if no match
    """
    # list_distributions returns at most 100 distributions per request
    list_limit = "100"
    distributions = EFAwsResolver.__CLIENTS["cloudfront"].list_distributions(MaxItems=list_limit)["DistributionList"]
    # Return if the account has no Distributions
    if "Items" not in distributions:
      return default
    while True:
      for distribution in distributions["Items"]:
        if lookup in distribution["Aliases"]["Items"]:
          return distribution["DomainName"]
      if distributions["IsTruncated"]:
        distributions = EFAwsResolver.__CLIENTS["cloudfront"].list_distributions(
          MaxItems=list_limit, Marker=distributions["NextMarker"])["DistributionList"]
      else:
        return default

  def cloudfront_origin_access_identity_oai_id(self, lookup, default=None):
    """
    Args:
      lookup: the FQDN of the Origin Access Identity (from its comments)
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the ID of the Origin Access Identity associated with the named FQDN in 'lookup', or default/None if no match
    """
    # list_cloud_front_origin_access_identities returns at most 100 oai's per request
    list_limit = "100"
    oais = EFAwsResolver.__CLIENTS["cloudfront"].list_cloud_front_origin_access_identities(
      MaxItems=list_limit)["CloudFrontOriginAccessIdentityList"]
    # Return if the account has no OriginAccessIdentities
    if "Items" not in oais:
      return default
    while True:
      for oai in oais["Items"]:
        if oai["Comment"] == lookup:
          return oai["Id"]
      if oais["IsTruncated"]:
        oais = EFAwsResolver.__CLIENTS["cloudfront"].list_cloud_front_origin_access_identities(
          MaxItems=list_limit, Marker=oais["NextMarker"])["CloudFrontOriginAccessIdentityList"]
      else:
        return default

  def cloudfront_origin_access_identity_oai_canonical_user_id(self, lookup, default=None):
    """
    Args:
      lookup: the FQDN of the Origin Access Identity (from its comments)
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      the S3 Canonical User ID of the OAI associated with the named FQDN in 'lookup', or default/None if no match
    """
    # list_cloud_front_origin_access_identities returns at most 100 oai's per request
    list_limit = "100"
    oais = EFAwsResolver.__CLIENTS["cloudfront"].list_cloud_front_origin_access_identities(
      MaxItems=list_limit)["CloudFrontOriginAccessIdentityList"]
    # Return if the account has no OriginAccessIdentities
    if "Items" not in oais:
      return default
    while True:
      for oai in oais["Items"]:
        if oai["Comment"] == lookup:
          return oai["S3CanonicalUserId"]
      if oais["IsTruncated"]:
        oais = EFAwsResolver.__CLIENTS["cloudfront"].list_cloud_front_origin_access_identities(
          MaxItems=list_limit, Marker=oais["NextMarker"])["CloudFrontOriginAccessIdentityList"]
      else:
        return default

  def cognito_identity_identity_pool_arn(self, lookup, default=None):
    """
    Args:
        lookup: Cognito Federated Identity name, proto0-cms-identity-pool
        default: the optional value to return if lookup failed; returns None if not set

    Returns:
        the constructed ARN for the cognito identity pool, else default/None
    """
    identity_pool_id = self.cognito_identity_identity_pool_id(lookup, default)

    if identity_pool_id == default:
      return default

    # The ARN has to be constructed because there is no boto3 call that returns the full ARN for a cognito identity pool
    return "arn:aws:cognito-identity:{{{{REGION}}}}:{{{{ACCOUNT}}}}:identitypool/{}".format(identity_pool_id)

  def cognito_identity_identity_pool_id(self, lookup, default=None):
    """
    Args:
        lookup: Cognito Federated Identity name, proto0-cms-identity-pool
        default: the optional value to return if lookup failed; returns None if not set

    Returns:
        the Cognito Identity Pool ID corresponding to the given lookup, else default/None
    """
    # List size cannot be greater than 60
    list_limit = 60
    client = EFAwsResolver.__CLIENTS["cognito-identity"]
    response = client.list_identity_pools(MaxResults=list_limit)

    while "IdentityPools" in response:
      # Loop through all the identity pools
      for pool in response["IdentityPools"]:
        if pool["IdentityPoolName"] == lookup:
          return pool["IdentityPoolId"]

      # No match found on this page, but there are more pages
      if "NextToken" in response:
        response = client.list_identity_pools(MaxResults=list_limit, NextToken=response["NextToken"])
      else:
        break

    return default

  def cognito_idp_user_pool_arn(self, lookup, default=None):
    """
    Args:
        lookup: Cognito User Pool name, proto0-cms-user-pool
        default: the optional value to return if lookup failed; returns None if not set

    Returns:
        the User Pool ARN corresponding to the given lookup, else default/None
    """
    client = EFAwsResolver.__CLIENTS["cognito-idp"]
    user_pool_id = self.cognito_idp_user_pool_id(lookup, default)
    if user_pool_id == default:
      return default

    response = client.describe_user_pool(UserPoolId=user_pool_id)

    if "UserPool" not in response:
      return default

    return response["UserPool"]["Arn"]

  def cognito_idp_user_pool_id(self, lookup, default=None):
    """
    Args:
        lookup: Cognito User Pool name, proto0-cms-user-pool
        default: the optional value to return if lookup failed; returns None if not set

    Returns:
        the User Pool ID corresponding to the given lookup, else default/None
    """
    # List size cannot be greater than 60
    list_limit = 60
    client = EFAwsResolver.__CLIENTS["cognito-idp"]
    response = client.list_user_pools(MaxResults=list_limit)

    while "UserPools" in response:
      # Loop through all user pools
      for pool in response["UserPools"]:
        if pool["Name"] == lookup:
          return pool["Id"]

      # No match found on this page, but there are more pages
      if "NextToken" in response:
        response = client.list_identity_pools(MaxResults=list_limit, NextToken=response["NextToken"])
      else:
        break

    return default

  def kms_decrypt_value(self, lookup):
    """
    Args:
      lookup: the encrypted value to be decrypted by KMS; base64 encoded
    Returns:
      The decrypted lookup value
    """
    decrypted_lookup = ef_utils.kms_decrypt(EFAwsResolver.__CLIENTS["kms"], lookup)
    return decrypted_lookup.plaintext.decode('string_escape')

  def kms_key_arn(self, lookup):
    """
    Args:
      lookup: The key alias, EX: alias/proto0-evs-drm
    Returns:
      The full key arn
    """
    key_arn = ef_utils.kms_key_arn(EFAwsResolver.__CLIENTS["kms"], lookup)
    return key_arn

  def ram_resource_share_arn(self, lookup, default=None):
    """
    Args:
      lookup: the name of the resource share to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The arn of the first resource share found with a label matching 'lookup' or default/None if no match found
    """
    resource_shares = EFAwsResolver.__CLIENTS["ram"].get_resource_shares(
      resourceOwner="OTHER-ACCOUNTS",
      name=lookup)

    if len(resource_shares.get("resourceShares")) > 0:
      return resource_shares["resourceShares"][0]["resourceShareArn"]
    else:
      return default

  def ram_resource_arn(self, lookup, default=None):
    """
    Args:
      lookup: the resource share arn to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The arn of the first resource found with a label matching 'lookup' or default/None if no match found
    """
    resources = EFAwsResolver.__CLIENTS["ram"].list_resources(
      resourceOwner="OTHER-ACCOUNTS",
      resourceShareArns=[lookup])

    if len(resources.get("resources")) > 0:
      return resources["resources"][0]["arn"]
    else:
      return default

  def ec2_transit_gateway_id(self, lookup, default=None):
    """
    Args:
      lookup: the arn of the transit gateway to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The id of the first transit gateway found with a label matching 'lookup' or default/None if no match found
    """
    transit_gateways = EFAwsResolver.__CLIENTS["ec2"].describe_transit_gateways()

    if len(transit_gateways.get("TransitGateways")) > 0:
      transit_gateways = [i for i in transit_gateways["TransitGateways"] if (i["TransitGatewayArn"] == lookup)]
      return transit_gateways[0]["TransitGatewayId"]
    else:
      return default

  def ecr_repository_uri(self, lookup, default=None):
    """
    Args:
      lookup: the name of the Docker image to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The id of the first image found with a label matching 'lookup' or default/None if no match found
    """
    try:
      repositories = EFAwsResolver.__CLIENTS["ecr"].describe_repositories(repositoryNames=[ lookup ])
      if len(repositories.get("repositories")) > 0:
        return repositories.get("repositories")[0]["repositoryUri"]
      else:
        return default
    except ClientError:
      return default

  def dynamodb_stream_arn(self, lookup, default=None):
    """
    Args:
      lookup: the name of the DynamoDB table to look up
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The DynamoDB Stream ARN with a label matching 'lookup' or default/None if no match found
    """
    try:
      dynamodb_table = EFAwsResolver.__CLIENTS["dynamodb"].describe_table(TableName=lookup)
      if dynamodb_table:
        return dynamodb_table["Table"]["LatestStreamArn"]
      else:
        return default
    except ClientError:
      return default

  def lookup(self, token):
    try:
      kv = token.split(",")
    except (ValueError, AttributeError):
      return None
    if kv[0] == "acm:certificate-arn":
      return self.acm_certificate_arn(*kv[1:])
    elif kv[0] == "cloudfront:domain-name":
      return self.cloudfront_domain_name(*kv[1:])
    elif kv[0] == "cloudfront:origin-access-identity/oai-canonical-user-id":
      return self.cloudfront_origin_access_identity_oai_canonical_user_id(*kv[1:])
    elif kv[0] == "cloudfront:origin-access-identity/oai-id":
      return self.cloudfront_origin_access_identity_oai_id(*kv[1:])
    elif kv[0] == "cognito-identity:identity-pool-arn":
      return self.cognito_identity_identity_pool_arn(*kv[1:])
    elif kv[0] == "cognito-identity:identity-pool-id":
      return self.cognito_identity_identity_pool_id(*kv[1:])
    elif kv[0] == "cognito-idp:user-pool-arn":
      return self.cognito_idp_user_pool_arn(*kv[1:])
    elif kv[0] == "cognito-idp:user-pool-id":
      return self.cognito_idp_user_pool_id(*kv[1:])
    elif kv[0] == "dynamodb:stream-arn":
      return self.dynamodb_stream_arn(*kv[1:])
    elif kv[0] == "ec2:elasticip/elasticip-id":
      return self.ec2_elasticip_elasticip_id(*kv[1:])
    elif kv[0] == "ec2:elasticip/elasticip-ipaddress":
      return self.ec2_elasticip_elasticip_ipaddress(*kv[1:])
    elif kv[0] == "ec2:eni/eni-id":
      return self.ec2_eni_eni_id(*kv[1:])
    elif kv[0] == "ec2:network/network-acl-id":
      return self.ec2_network_network_acl_id(*kv[1:])
    elif kv[0] == "ec2:route-table/main-route-table-id":
      return self.ec2_route_table_main_route_table_id(*kv[1:])
    elif kv[0] == "ec2:route-table/tagged-route-table-id":
      return self.ec2_route_table_tagged_route_table_id(*kv[1:])
    elif kv[0] == "ec2:security-group/security-group-id":
      return self.ec2_security_group_security_group_id(*kv[1:])
    elif kv[0] == "ec2:subnet/subnet-cidr":
      return self.ec2_subnet_subnet_cidr(*kv[1:])
    elif kv[0] == "ec2:subnet/subnet-id":
      return self.ec2_subnet_subnet_id(*kv[1:])
    elif kv[0] == "ec2:transit-gateway/transit-gateway-id":
      return self.ec2_transit_gateway_id(*kv[1:])
    elif kv[0] == "ec2:vpc/availabilityzones":
      return self.ec2_vpc_availabilityzones(*kv[1:])
    elif kv[0] == "ec2:vpc/cidrblock":
      return self.ec2_vpc_cidrblock(*kv[1:])
    elif kv[0] == "ec2:vpc/subnets":
      return self.ec2_vpc_subnets(*kv[1:])
    elif kv[0] == "ec2:vpc/vpc-id":
      return self.ec2_vpc_vpc_id(*kv[1:])
    elif kv[0] == "ec2:vpc-endpoint/vpc-endpoint-id":
      return self.ec2_vpc_endpoint_id(*kv[1:])
    elif kv[0] == "ec2:vpc-endpoint/vpc-endpoint-id/by-vpc-service":
      return self.ec2_vpc_endpoint_id_by_vpc_service(*kv[1:])
    elif kv[0] == "ec2:vpc-endpoint/dns-name":
      return self.ec2_vpc_endpoint_dns_name(*kv[1:])
    elif kv[0] == "ec2:vpc-endpoint/dns-name/by-vpc-service":
      return self.ec2_vpc_endpoint_dns_name_by_vpc_service(*kv[1:])
    elif kv[0] == "ec2:vpc/vpn-gateway-id":
      return self.ec2_vpc_vpn_gateway_id(*kv[1:])
    elif kv[0] == "ecr:repository/repository-uri":
      return self.ecr_repository_uri(*kv[1:])
    elif kv[0] == "elbv2:load-balancer/dns-name":
      return self.elbv2_load_balancer_dns_name(*kv[1:])
    elif kv[0] == "elbv2:load-balancer/hosted-zone":
      return self.elbv2_load_balancer_hosted_zone(*kv[1:])
    elif kv[0] == "elbv2:load-balancer/arn-suffix":
      return self.elbv2_load_balancer_arn_suffix(*kv[1:])
    elif kv[0] == "elbv2:target-group/arn-suffix":
      return self.elbv2_target_group_arn_suffix(*kv[1:])
    elif kv[0] == "kms:decrypt":
      return self.kms_decrypt_value(*kv[1:])
    elif kv[0] == "kms:key_arn":
      return self.kms_key_arn(*kv[1:])
    elif kv[0] == "ram:resource-share/resource-share-arn":
      return self.ram_resource_share_arn(*kv[1:])
    elif kv[0] == "ram:resource-share/resource-arn":
      return self.ram_resource_arn(*kv[1:])
    elif kv[0] == "route53:private-hosted-zone-id":
      return self.route53_private_hosted_zone_id(*kv[1:])
    elif kv[0] == "route53:public-hosted-zone-id":
      return self.route53_public_hosted_zone_id(*kv[1:])
    elif kv[0] == "waf:ip-set-id":
      return self.waf_ip_set_id(*kv[1:])
    elif kv[0] == "waf:rule-id":
      return self.waf_rule_id(*kv[1:])
    elif kv[0] == "waf:web-acl-id":
      return self.waf_web_acl_id(*kv[1:])
    else:
      return None
      # raise("No lookup function for: "+kv[0])


  def __init__(self, clients):
    """
    ARGS
      clients - dictionary of ready-to-go boto3 clients using aws prefixes:
      expected: clients["ec2"], clients["iam"], clients["lambda"]
    """
    EFAwsResolver.__CLIENTS = clients



"""
provides simple utility functions used in many scripts

Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
import base64
from collections import namedtuple
import json
from os import access, getenv, X_OK
from os.path import isfile
import re
from socket import gethostname
import subprocess
import sys
import urllib2

import boto3
from botocore.exceptions import ClientError

__HTTP_DEFAULT_TIMEOUT_SEC = 5
__METADATA_PREFIX = "http://169.254.169.254/latest/meta-data/"
__VIRT_WHAT = "/sbin/virt-what"
__VIRT_WHAT_VIRTUALBOX_WITH_KVM = ["virtualbox", "kvm"]

# Matches CIDRs (loosely, TBH)
CIDR_REGEX = r"^(([1-2][0-9]{2}|[0-9]{0,2})\.){3}([1-2][0-9]{2}|[0-9]{0,2})\/([1-3][0-9]|[0-9])$"

# Cache for AWS clients. Keeps all the clients under (region, profile) keys.
client_cache = {}

DecryptedSecret = namedtuple("DecryptedSecret", ["plaintext", "key_id"])


def fail(message, exception_data=None):
  """
  Print a failure message and exit nonzero
  """
  print(message, file=sys.stderr)
  if exception_data:
    print(repr(exception_data))
  sys.exit(1)

def http_get_metadata(metadata_path, timeout=__HTTP_DEFAULT_TIMEOUT_SEC):
  """
  Fetch AWS metadata from http://169.254.169.254/latest/meta-data/<metadata_path>
  ARGS:
    metadata_path - the optional path and required key to the EC2 metadata (e.g. "instance-id")
  RETURN:
    response content on success
  RAISE:
    URLError if there was a problem reading metadata
  """
  metadata_path = __METADATA_PREFIX + metadata_path
  try:
    response = urllib2.urlopen(metadata_path, None, timeout)
    if response.getcode() != 200:
      raise IOError("Non-200 response " + str(response.getcode()) + " reading " + metadata_path)
    return response.read()
  except urllib2.URLError as error:
    raise IOError("URLError in http_get_metadata: " + repr(error))

def is_in_virtualbox():
  """
  Is the current environment a virtualbox instance?
  Returns a boolean
  Raises IOError if the necessary tooling isn't available
  """
  if not isfile(__VIRT_WHAT) or not access(__VIRT_WHAT, X_OK):
    raise IOError("virt-what not available")
  try:
    return subprocess.check_output(["sudo", "-n", __VIRT_WHAT]).split('\n')[0:2] == __VIRT_WHAT_VIRTUALBOX_WITH_KVM
  except subprocess.CalledProcessError as e:
    raise IOError("virt-what failed execution with {}".format(e))

def whereami():
  """
  Determine if this is an ec2 instance or "running locally"
  Returns:
    "ec2" - this is an ec2 instance
    "jenkins" - running inside a Jenkins job
    "virtualbox-kvm" - kernel VM (virtualbox with vagrant)
    "local" - running locally and not in a known VM
    "unknown" - I have no idea where I am
  """
  if getenv("JENKINS_URL") and getenv("JENKINS_DOCKER") is None:
    # The addition of the JENKINS_DOCKER is a temporary workaround to have Jenkins Docker machine rely on its instance
    # role and assume roles vs a credentials file. This is an on-going effort to move everything to code with
    # https://ellation.atlassian.net/browse/OPS-13637
    # Regular Jenkins machines will still continue to use their credentials files until we switch over.
    return "jenkins"

  # If the metadata endpoint responds, this is an EC2 instance
  # If it doesn't, we can safely say this isn't EC2 and try the other options
  try:
    response = http_get_metadata("instance-id", 1)
    if response[:2] == "i-":
      return "ec2"
  except:
    pass

  # Virtualbox?
  try:
    if is_in_virtualbox():
      return "virtualbox-kvm"
  except:
    pass

  # Outside virtualbox/vagrant but not in aws; hostname is "<name>.local"
  hostname = gethostname()
  if re.findall(r"\.local$", hostname):
    return "local"

  # we have no idea where we are
  return "unknown"

def http_get_instance_env():
  """
  Returns: just the env this ec2 instance is in. Doesn't require API access like get_instance_aws_context does
  Example return value: "staging"
  """
  try:
    info = json.loads(http_get_metadata('iam/info'))
  except Exception as error:
    raise IOError("Error looking up metadata:iam/info: " + repr(error))
  return info["InstanceProfileArn"].split(":")[5].split("/")[1].split("-",1)[0]

def http_get_instance_role():
  """
  Returns: just the role this ec2 instance is in. Doesn't require API access like get_instance_aws_context does
  Example return: "vrvweb"
  """
  try:
    info = json.loads(http_get_metadata('iam/info'))
  except Exception as error:
    raise IOError("Error looking up metadata:iam/info: " + repr(error))
  return info["InstanceProfileArn"].split(":")[5].split("/")[1].split("-",1)[1]

def create_aws_clients(region, profile, *clients):
  """
  Create boto3 clients for one or more AWS services. These are the services used within the libs:
    cloudformation, cloudfront, ec2, iam, lambda, route53, waf
  Args:
    region: the region in which to create clients that are region-specific (all but IAM)
    profile: Name of profile (in .aws/credentials). Pass the value None if using instance credentials on EC2 or Lambda
    clients: names of the clients to create (lowercase, must match what boto3 expects)
  Returns:
    A dictionary of <key>,<value> pairs for several AWS services, using the labels above as keys, e.g.:
    { "cloudfront": <cloudfront_client>, ... }
    Dictionary contains an extra record, "SESSION" - pointing to the session that created the clients
  """
  if not profile:
    profile = None

  client_key = (region, profile)

  aws_clients = client_cache.get(client_key, {})
  requested_clients = set(clients)
  new_clients = requested_clients.difference(aws_clients)

  if not new_clients:
    return aws_clients

  session = aws_clients.get("SESSION")
  try:
    if not session:
      session = boto3.Session(region_name=region, profile_name=profile)
      aws_clients["SESSION"] = session
    # build clients
    client_dict = {c: session.client(c) for c in new_clients}
    # append the session itself in case it's needed by the client code - can't get it from the clients themselves
    aws_clients.update(client_dict)

    # add the created clients to the cache
    client_cache[client_key] = aws_clients
    return aws_clients
  except ClientError as error:
    raise RuntimeError("Exception logging in with Session() and creating clients", error)

def get_account_id(sts_client):
  """
  Args:
    sts_client (boto3 sts client object): Instantiated sts client object. Usually created through create_aws_clients
  """
  return sts_client.get_caller_identity().get('Account')

def kms_encrypt(kms_client, service, env, secret):
  """
  Encrypt string for use by a given service/environment
  Args:
    kms_client (boto3 kms client object): Instantiated kms client object. Usually created through create_aws_clients.
    service (string): name of the service that the secret is being encrypted for.
    env (string): environment that the secret is being encrypted for.
    secret (string): value to be encrypted
  Returns:
    a populated EFPWContext object
  Raises:
    SystemExit(1): If there is an error with the boto3 encryption call (ex. missing kms key)
  """
  # Converting all periods to underscores because they are invalid in KMS alias names
  key_alias = '{}-{}'.format(env, service.replace('.', '_'))

  try:
    response = kms_client.encrypt(
      KeyId='alias/{}'.format(key_alias),
      Plaintext=secret.encode()
    )
  except ClientError as error:
    if error.response['Error']['Code'] == "NotFoundException":
      fail("Key '{}' not found. You may need to run ef-generate for this environment.".format(key_alias), error)
    else:
      fail("boto3 exception occurred while performing kms encrypt operation.", error)
  encrypted_secret = base64.b64encode(response['CiphertextBlob'])
  return encrypted_secret

def kms_decrypt(kms_client, secret):
  """
  Decrypt kms-encrypted string
  Args:
    kms_client (boto3 kms client object): Instantiated kms client object. Usually created through create_aws_clients.
    secret (string): base64 encoded value to be decrypted
  Returns:
    DecryptedSecret object
  Raises:
    SystemExit(1): If there is an error with the boto3 decryption call (ex. malformed secret)
  """
  try:
    response = kms_client.decrypt(CiphertextBlob=base64.b64decode(secret))
    decrypted_secret = DecryptedSecret(response["Plaintext"], response["KeyId"])
  except TypeError as e:
    fail("Malformed base64 string data: {}".format(e))
  except ClientError as error:
    if error.response["Error"]["Code"] == "InvalidCiphertextException":
      fail("The decrypt request was rejected because the specified ciphertext \
      has been corrupted or is otherwise invalid.", error)
    elif error.response["Error"]["Code"] == "NotFoundException":
      fail("The decrypt request was rejected because the specified entity or resource could not be found.", error)
    else:
      fail("boto3 exception occurred while performing kms decrypt operation.", error)
  return decrypted_secret

def kms_re_encrypt(kms_client, service, env, secret):
  """
  Re-encrypt the secret for a new service. Don't need to know the plaintext
  Args:
    kms_client (boto3 kms client object): Instantiated kms client object. Usually created through create_aws_clients.
    service (string): name of the service that the secret is being encrypted for.
    env (string): environment that the secret is being encrypted for.
    secret (string): base64 encoded value to be reencrypted
  Returns:
    a populated EFPWContext object
  Raises:
    SystemExit(1): If there is an error with the boto3 encryption call (ex. missing kms key)
  """
  # Converting all periods to underscores because they are invalid in KMS alias names
  key_alias = '{}-{}'.format(env, service.replace('.', '_'))

  try:
    response = kms_client.re_encrypt(
      DestinationKeyId='alias/{}'.format(key_alias),
      CiphertextBlob=base64.b64decode(secret)
    )
  except TypeError as e:
    fail("Malformed base64 string data: {}".format(e))
  except ClientError as error:
    if error.response['Error']['Code'] == "NotFoundException":
      fail("Key '{}' not found. You may need to run ef-generate for this environment.".format(key_alias), error)
    else:
      fail("boto3 exception occurred while performing kms encrypt operation.", error)
  encrypted_secret = base64.b64encode(response['CiphertextBlob'])
  return encrypted_secret

def kms_key_alias(kms_client, key_arn):
  """
  Obtain the key aliases based on the key arn provided
  Args:
    kms_client (boto3 kms client object): Instantiated kms client object. Usually created through create_aws_clients.
    key_arn (string): key arn

  Returns:
    list of aliases associated with the key
  """
  try:
    response = kms_client.list_aliases(KeyId=key_arn)
    key_aliases = [key_data["AliasName"] for key_data in response["Aliases"]]
    clean_aliases = [alias.split('/', 1)[1] for alias in key_aliases]
  except ClientError as error:
    raise RuntimeError("Failed to obtain key alias for arn {}, error: {}".format(key_arn, error.response["Error"]["Message"]))

  return clean_aliases

def kms_key_arn(kms_client, alias):
  """
  Obtain the full key arn based on the key alias provided
  Args:
    kms_client (boto3 kms client object): Instantiated kms client object. Usually created through create_aws_clients.
    alias (string): alias of key, example alias/proto0-evs-drm.

  Returns:
    string of the full key arn
  """
  try:
    response = kms_client.describe_key(KeyId=alias)
    key_arn = response["KeyMetadata"]["Arn"]
  except ClientError as error:
    raise RuntimeError("Failed to obtain key arn for alias {}, error: {}".format(alias, error.response["Error"]["Message"]))

  return key_arn

def get_autoscaling_group_properties(asg_client, env, service):
  """
  Gets the autoscaling group properties based on the service name that is provided. This function will attempt the find
  the autoscaling group base on the following logic:
    1. If the service name provided matches the autoscaling group name
    2. If the service name provided matches the Name tag of the autoscaling group
    3. If the service name provided does not match the above, return None
  Args:
    clients: Instantiated boto3 autoscaling client
    env: Name of the environment to search for the autoscaling group
    service: Name of the service
  Returns:
    JSON object of the autoscaling group properties if it exists
  """
  try:
    # See if {{ENV}}-{{SERVICE}} matches ASG name
    response = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=["{}-{}".format(env, service)])
    if len(response["AutoScalingGroups"]) == 0:
      # See if {{ENV}}-{{SERVICE}} matches ASG tag name
      response = asg_client.describe_tags(Filters=[{ "Name": "Key", "Values": ["Name"] }, { "Name": "Value", "Values": ["{}-{}".format(env, service)]}])
      if len(response["Tags"]) == 0:
        # Query does not match either of the above, return None
        return None
      else:
         asg_name = response["Tags"][0]["ResourceId"]
         response = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])
         return response["AutoScalingGroups"]
    else:
      return response["AutoScalingGroups"]
  except ClientError as error:
    raise RuntimeError("Error in finding autoscaling group {} {}".format(env, service), error)



#!/usr/bin/env python
"""
Validates service json configuration templates and/or Cloudformation parameter files

Exit codes:
  0 if all tested configs were valid
  1 if one or more tested configs had errors
  2 if an I/O error occurred when processing a file
  3 if some other error occurred


Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
import argparse
import json
from os import walk
import os.path
import sys

CONFIGS_RELATIVE_PATH_FROM_SCRIPT_DIR = "../configs"
PARAMETER_SUFFIX = ".parameters.json"
DEFAULT_LOGFILE = "/tmp/instanceinit.log"

def handle_args(args):
  """
  Handle command line arguments
  Raises:
    Exception if the config path wasn't explicitly state and dead reckoning based on script location fails
  """
  parser = argparse.ArgumentParser(description="Check if a late bind config is valid or not.")
  parser.add_argument("configpath", default=None, nargs="?",
                      help="/path/to/configs (always a directory; if omitted, all /configs are checked")
  parser.add_argument("--verbose", action="store_true", default=False)
  parsed_args = vars(parser.parse_args(args))
  # If a config path wasn't given, calculate it based on location of the script
  if parsed_args["configpath"] is None:
    script_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
    parsed_args["configpath"] = os.path.normpath("{}/{}".format(script_dir, CONFIGS_RELATIVE_PATH_FROM_SCRIPT_DIR))
  else:
    parsed_args["configpath"] = os.path.normpath(parsed_args["configpath"])
  # If the path is a directory, all good. if it's a file, find the directory the file is in and check that instead
  if os.path.isdir(parsed_args["configpath"]):
    parsed_args["configdir"] = parsed_args["configpath"]
  else:
    parsed_args["configdir"] = os.path.dirname(parsed_args["configpath"])
  return parsed_args

def load_json(json_filespec):
  """
  Loads JSON from a config file
  Args:
    json_filespec: path/to/file.json
  Returns:
    a dict made from the JSON read, if successful
  Raises:
    IOError if the file could not be opened
    ValueError if the JSON could not be read successfully
    RuntimeError if something else went wrong
  """
  json_fh = open(json_filespec)
  config_dict = json.load(json_fh)
  json_fh.close()
  return config_dict

def check_file(json_filespec):
  load_json(json_filespec)


def main():
  args = handle_args(sys.argv[1:])
  if args["verbose"]:
    print("Path: {}".format(args["configpath"]))
    print("Dir: {}".format(args["configdir"]))

  # Confirm that all parameters files can be loaded
  exit_code = 0
  for root, dirs, files in walk(args["configdir"]):
    for f in files:
      filename = os.path.join(root, f)
      # dot-files aren't checked
      if filename.startswith("."):
        continue
      # Only 'parameters' files, not templates
      if not os.path.dirname(filename).endswith("/parameters"):
        continue
      if args["verbose"]:
        print(filename)
      try:
        check_file(filename)
      except ValueError:
        print("JSON syntax error: {}".format(filename))
        exit_code = 1
      except IOError as e:
        print("IO error processing file:\n{}\n{}".format(filename, e))
        sys.exit(2)
      except Exception as e:
        print("Exception processing file:\n{}\n{}".format(filename, e))
        sys.exit(3)

  sys.exit(exit_code)

if __name__ == "__main__":
  main()



#!/usr/bin/env python

"""
For one environment:
- Step through all services in the service registry
- If a security group and/or role named "<env>-<service>" is needed in the current account and
environment, make it.
- Make EC2 instance profiles as needed as well.
- Bind policies from policy_templates to roles if "policies" key is present.
- Override default AssumeRole policy if assume_role_policy key is present.

If --commit flag is not set, only a dry run occurs; no changes are made.


Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
import argparse
import json
from os.path import dirname, normpath
import sys
import time

from botocore.exceptions import ClientError

from ef_aws_resolver import EFAwsResolver
from ef_config import EFConfig
from ef_context import EFContext
from ef_service_registry import EFServiceRegistry
from ef_template_resolver import EFTemplateResolver
from ef_utils import create_aws_clients, fail, get_account_id, http_get_metadata
from ef_conf_utils import pull_repo

# Globals
CLIENTS = None
CONTEXT = None
AWS_RESOLVER = None

# list of service_registry_file "service types" that this tool handles. Skip unlisted types
SUPPORTED_SERVICE_TYPES = [
  "aws_cloudtrail",
  "aws_ec2",
  "aws_ecs",
  "aws_ecs_http",
  "aws_fixture",
  "aws_lambda",
  "aws_role",
  "aws_security_group",
  "http_service"
]

# list of service_registry_file "service types" that are allowed in the global env
GLOBAL_SERVICE_TYPES = [
  "aws_cloudtrail",
  "aws_fixture",
  "aws_lambda",
  "aws_role"
]

# these Service Registry types get Roles
# If value is not None, the role can get a default AssumeRolePolicy document listing that AWS service as principal
# If value is None, the service registry entry must include "assume_role_policy": with an AssumeRole policy doc
SERVICE_TYPE_ROLE = {
  "aws_cloudtrail": "cloudtrail.amazonaws.com",
  "aws_ec2": "ec2.amazonaws.com",
  "aws_ecs": ["ecs.amazonaws.com", "ecs-tasks.amazonaws.com"],
  "aws_ecs_http": ["ecs.amazonaws.com", "ecs-tasks.amazonaws.com"],
  "aws_lambda": "lambda.amazonaws.com",
  "aws_role": None, # Bare role requires a custom assume role policy ("assume_role_policy": "name_of_policy")
  "http_service": "ec2.amazonaws.com"
}

# these service types get Security Groups
SG_SERVICE_TYPES = [
  "aws_ec2",
  "aws_ecs",
  "aws_ecs_http",
  "aws_lambda",
  "aws_security_group",
  "http_service"
]

# these service types get instance profiles (they have EC2 instances)
INSTANCE_PROFILE_SERVICE_TYPES = [
  "aws_ec2",
  "aws_ecs",
  "aws_ecs_http",
  "http_service"
]

# these service types get KMS Keys
KMS_SERVICE_TYPES = [
  "aws_ec2",
  "aws_ecs",
  "aws_ecs_http",
  "aws_fixture",
  "aws_lambda",
  "http_service"
]

# Utilities
def handle_args_and_set_context(args):
  """
  Args:
    args: the command line args, probably passed from main() as sys.argv[1:]
  Returns:
    a populated EFContext object
  Raises:
    IOError: if service registry file can't be found or can't be opened
    RuntimeError: if branch isn't as spec'd in ef_config.EF_REPO_BRANCH
    CalledProcessError: if 'git rev-parse' command to find repo root could not be run
  """
  parser = argparse.ArgumentParser(
    description="Generate a security group, IAM role, instance profile, attache managed, custom, or inline "
                "generated policies to a role, create KMS keys, and create newrelic alerts for a service based "
                "on the information specified in the service_registry.json.")
  parser.add_argument("env", help=", ".join(EFConfig.ENV_LIST))
  parser.add_argument("--sr", help="optional: /path/to/service_registry_file.json", default=None)
  parser.add_argument("--commit", help="Make changes in AWS (dry run if omitted)", action="store_true", default=False)
  parser.add_argument("--verbose", help="Print additional info", action="store_true", default=False)
  parser.add_argument("--devel", help="Allow running from branch; don't refresh from origin", action="store_true",
                      default=False)
  parser.add_argument("--only_for", help="optional: run ef-generate for a only specific entry in the service registry", default=None)
  parsed_args = vars(parser.parse_args(args))
  context = EFContext()
  context.commit = parsed_args["commit"]
  context.devel = parsed_args["devel"]
  try:
    context.env = parsed_args["env"]
  except ValueError as e:
    fail("Error in env: {}".format(e))
  # Set up service registry and policy template path which depends on it
  context.service_registry = EFServiceRegistry(parsed_args["sr"])
  context.policy_template_path = normpath(dirname(context.service_registry.filespec)) + EFConfig.POLICY_TEMPLATE_PATH_SUFFIX
  context.verbose = parsed_args["verbose"]
  context.only_for = parsed_args["only_for"]
  return context

def print_if_verbose(message):
  if CONTEXT.verbose:
    print(message, file=sys.stderr)

def get_role_id(target_name):
  """
  Args:
    target_name: the name of the role to look up
  Returns:
    role ID if target_name found, False otherwise
  """
  try:
    role = CLIENTS["iam"].get_role(RoleName=target_name)
  except:
    return False
  return role["Role"]["RoleId"]

def get_instance_profile(instance_profile_name):
  """
  Args:
    instance_profile_name: the name of the instance profile to look up
  Returns:
    an instance profile if target_name found, False otherwise
  """
  try:
    instance_profile = CLIENTS["iam"].get_instance_profile(InstanceProfileName=instance_profile_name)
  except:
    return False
  return instance_profile

def instance_profile_contains_role(instance_profile, role_name):
  """
  Args:
    instance_profile: the instance profile to check
    role_name: the name of the role to look for
  Returns:
    True if role_name is inside instance_profile, False otherwise
  """
  for role in instance_profile["InstanceProfile"]["Roles"]:
    if role["RoleName"] == role_name:
      return True
    return False

def resolve_policy_document(policy_name):
  policy_filename = "{}{}.json".format(CONTEXT.policy_template_path, policy_name)
  print_if_verbose("Load policy: {} from file: {}".format(policy_name, policy_filename))
  # retrieve policy template
  try:
    policy_file = file(policy_filename, 'r')
    policy_template = policy_file.read()
    policy_file.close()
  except:
    fail("error opening policy file: {}".format(policy_filename))
  print_if_verbose("pre-resolution policy template:\n{}".format(policy_template))
  # If running in EC2, do not set profile and set target_other=True
  if CONTEXT.whereami == "ec2":
    resolver = EFTemplateResolver(target_other=True, env=CONTEXT.env, region=EFConfig.DEFAULT_REGION,
                                  service=CONTEXT.service, verbose=CONTEXT.verbose)
  else:
    resolver = EFTemplateResolver(profile=CONTEXT.account_alias, env=CONTEXT.env, region=EFConfig.DEFAULT_REGION,
                                service=CONTEXT.service, verbose=CONTEXT.verbose)
  resolver.load(policy_template)
  policy_document = resolver.render()
  print_if_verbose("resolved policy document:\n{}".format(policy_document))
  if not resolver.resolved_ok():
    fail("policy template {} has unresolved symbols or extra {{ or }}: {}".format(
      policy_filename, resolver.unresolved_symbols()))
  return policy_document

def conditionally_create_security_groups(env, service_name, service_type):
  """
  Create security groups as needed; name and number created depend on service_type
  Args:
    env: the environment the SG will be created in
    service_name: name of the service in service registry
    service_type: service registry service type: 'aws_ec2', 'aws_ecs', 'aws_ecs_http',
                                 'aws_lambda', 'aws_security_group', or 'http_service'
  """
  if service_type not in SG_SERVICE_TYPES:
    print_if_verbose("not eligible for security group(s); service type: {}".format(service_type))
    return

  target_name = "{}-{}".format(env, service_name)
  if service_type == "aws_ec2":
    sg_names = ["{}-ec2".format(target_name)]
  elif service_type == "aws_ecs":
    sg_names = ["{}-ecs".format(target_name)]
  elif service_type == "aws_lambda":
    sg_names = ["{}-lambda".format(target_name)]
  elif service_type == "http_service":
    sg_names = [
      "{}-ec2".format(target_name),
      "{}-elb".format(target_name)
    ]
  elif service_type == "aws_ecs_http":
    sg_names = [
      "{}-ecs".format(target_name),
      "{}-elb".format(target_name)
    ]
  elif service_type == "aws_security_group":
    sg_names = [target_name]
  else:
    fail("Unexpected service_type: {} when creating security group for: {}".format(service_type, target_name))

  for sg_name in sg_names:
    if not AWS_RESOLVER.ec2_security_group_security_group_id(sg_name):
      vpc_name = "vpc-{}".format(env)
      print("Create security group: {} in vpc: {}".format(sg_name, vpc_name))
      vpc = AWS_RESOLVER.ec2_vpc_vpc_id(vpc_name)
      if not vpc:
        fail("Error: could not get VPC by name: {}".format(vpc_name))
      # create security group
      if CONTEXT.commit:
        try:
          new_sg = CLIENTS["ec2"].create_security_group(GroupName=sg_name, VpcId=vpc, Description=sg_name)
        except:
          fail("Exception creating security group named: {} in VpcId: {}".format(sg_name, vpc_name), sys.exc_info())
        print(new_sg["GroupId"])
    else:
      print_if_verbose("security group already exists: {}".format(sg_name))

def conditionally_create_role(role_name, sr_entry, tags=None):
  """
  Create role_name if a role by that name does not already exist; attach a custom list of Principals
  to its AssumeRolePolicy
  Args:
    role_name: the name for the role to create
    sr_entry: service registry entry

  Example of a (complex) AssumeRole policy document comprised of two IAM entities and a service:
  {
    "Version": "2012-10-17",
    "Statement": [
      {
        "Effect": "Allow",
        "Principal": {
          "Service": "ec2.amazonaws.com",
          "AWS": [
            "arn:aws:iam::978969509086:root",
            "arn:aws:iam::978969509086:role/mgmt-jenkins"
          ]
        },
        "Action": "sts:AssumeRole"
      }
    ]
  }
  """
  service_type = sr_entry['type']
  if service_type not in SERVICE_TYPE_ROLE:
    print_if_verbose("not eligible for role (and possibly instance profile); service type: {}".format(service_type))
    return

  if "assume_role_policy" in sr_entry:
    # Explicitly defined AssumeRole policy
    assume_role_policy_document = resolve_policy_document(sr_entry["assume_role_policy"])
  else:
    # Create Service:AssumeRole policy using the service type in the SERVICE_TYPE_ROLE dict
    # which must list a service type to use this capacity (most do)
    if SERVICE_TYPE_ROLE[service_type] is None:
      fail("service_type: {} does not have a default service-type AssumeRole policy".format(service_type))
    service_principals = {"Service": SERVICE_TYPE_ROLE[service_type]}
    assume_role_policy = {
      "Version" : "2012-10-17",
      "Statement": [{
        "Effect": "Allow",
        "Principal": service_principals,
        "Action": [ "sts:AssumeRole" ]
      }]
    }
    assume_role_policy_document = json.dumps(assume_role_policy, indent=2)
  if not get_role_id(role_name):
    print("Create role: {}".format(role_name))
    print_if_verbose("AssumeRole policy document:\n{}".format(assume_role_policy_document))
    if CONTEXT.commit:
      try:
        new_role = CLIENTS["iam"].create_role(
          RoleName=role_name, AssumeRolePolicyDocument=assume_role_policy_document, Tags=tags or []
        )
      except ClientError as error:
        fail("Exception creating new role named: {} {}".format(role_name, sys.exc_info(), error))
      print(new_role["Role"]["RoleId"])
  else:
    print_if_verbose("role already exists: {}".format(role_name))

def conditionally_create_profile(role_name, service_type):
  """
  Check that there is a 1:1 correspondence with an InstanceProfile having the same name
  as the role, and that the role is contained in it. Create InstanceProfile and attach to role if needed.
  """
  # make instance profile if this service_type gets an instance profile
  if service_type not in INSTANCE_PROFILE_SERVICE_TYPES:
    print_if_verbose("service type: {} not eligible for instance profile".format(service_type))
    return

  instance_profile = get_instance_profile(role_name)
  if not instance_profile:
    print("Create instance profile: {}".format(role_name))
    if CONTEXT.commit:
      try:
        instance_profile = CLIENTS["iam"].create_instance_profile(InstanceProfileName=role_name)
      except ClientError as error:
        fail("Exception creating instance profile named: {} {}".format(role_name, sys.exc_info(), error))
  else:
    print_if_verbose("instance profile already exists: {}".format(role_name))
  # attach instance profile to role; test 'if instance_profile' because we drop through to here in a dry run
  if instance_profile and not instance_profile_contains_role(instance_profile, role_name):
    print("Add role: {} to instance profile: {}".format(role_name, role_name))
    if CONTEXT.commit:
      try:
        CLIENTS["iam"].add_role_to_instance_profile(InstanceProfileName=role_name, RoleName=role_name)
      except ClientError as error:
        fail("Exception adding role to instance profile: {} {}".format(role_name, sys.exc_info(), error))
  else:
    print_if_verbose("instance profile already contains role: {}".format(role_name))


def get_managed_policy_arn(policy_name):
  """
  Retrieves the ARN of an AWS managed policy specified in the aws_managed_policies field of the service_registry.
  Args:
    policy_name: name of the policy whose ARN we want to retrieve
  Raises:
    NameError: no ARN could be found for the specified policy name
  """
  try:
    arn = 'arn:aws:iam::aws:policy/{}'.format(policy_name)
    CLIENTS["iam"].get_policy(PolicyArn=arn)
    return arn
  except CLIENTS["iam"].exceptions.NoSuchEntityException:
    pass

  role_types = ['job-function', 'service-role']
  for type in role_types:
    arn = 'arn:aws:iam::aws:policy/{}/{}'.format(type, policy_name)
    try:
      CLIENTS["iam"].get_policy(PolicyArn=arn)
      return arn
    except CLIENTS["iam"].exceptions.NoSuchEntityException:
      continue

  raise NameError


def conditionally_attach_aws_managed_policies(role_name, sr_entry):
  """
  If 'aws_managed_policies' key lists the names of AWS managed policies to bind to the role,
  attach them to the role
  Args:
    role_name: name of the role to attach the policies to
    sr_entry: service registry entry
  """
  service_type = sr_entry['type']
  if not (service_type in SERVICE_TYPE_ROLE and "aws_managed_policies" in sr_entry):
    print_if_verbose("not eligible for policies; service_type: {} is not valid for policies "
                     "or no 'aws_managed_policies' key in service registry for this role".format(service_type))
    return

  for policy_name in sr_entry['aws_managed_policies']:
    print_if_verbose("loading policy: {} for role: {}".format(policy_name, role_name))

    if CONTEXT.commit:
      try:
        policy_arn = get_managed_policy_arn(policy_name)
      except NameError:
        fail("Exception attaching managed policy '{}'. Could not retrieve policy arn.".format(policy_name))

      try:
        CLIENTS["iam"].attach_role_policy(RoleName=role_name, PolicyArn=policy_arn)
      except:
        fail("Exception putting policy: {} onto role: {}".format(policy_name, role_name), sys.exc_info())

      print_if_verbose("Attached managed policy '{}'".format(policy_name))


def conditionally_attach_customer_managed_policies(role_name, sr_entry):
  """
  If 'customer_managed_policies' key lists the names of customer managed policies to bind to the role,
  attach them to the role.

  Note that this function will throw a warning without failing, if the managed policy does not exist in
  the given account.
  Args:
    role_name: name of the role to attach the policies to
    sr_entry: service registry entry
  """
  service_type = sr_entry['type']
  if not (service_type in SERVICE_TYPE_ROLE and "customer_managed_policies" in sr_entry):
    print_if_verbose("not eligible for policies; service_type: {} is not valid for policies "
                     "or no 'customer_managed_policies' key in service registry for this role".format(service_type))
    return

  for policy_name in sr_entry['customer_managed_policies']:
    print_if_verbose("loading policy: {} for role: {}".format(policy_name, role_name))

    if CONTEXT.commit:
      policy_arn = 'arn:aws:iam::{}:policy/{}'.format(CONTEXT.account_id, policy_name)
      try:
        CLIENTS["iam"].attach_role_policy(RoleName=role_name, PolicyArn=policy_arn)
      except CLIENTS["iam"].exceptions.NoSuchEntityException as exc:
        print("WARNING: {}".format(exc))
      except:
        fail("Exception putting policy: {} onto role: {}".format(policy_name, role_name), sys.exc_info())

def conditionally_inline_policies(role_name, sr_entry):
  """
  If 'policies' key lists the filename prefixes of policies to bind to the role,
  load them from the expected path and inline them onto the role
  Args:
    role_name: name of the role to attach the policies to
    sr_entry: service registry entry
  """
  service_type = sr_entry['type']
  if not (service_type in SERVICE_TYPE_ROLE and "policies" in sr_entry):
    print_if_verbose("not eligible for policies; service_type: {} is not valid for policies "
                     "or no 'policies' key in service registry for this role".format(service_type))
    return

  for policy_name in sr_entry['policies']:
    print_if_verbose("loading policy: {} for role: {}".format(policy_name, role_name))
    try:
      policy_document = resolve_policy_document(policy_name)
    except:
      fail("Exception loading policy: {} for role: {}".format(policy_name, role_name), sys.exc_info())

    # inline the policy onto the role
    if CONTEXT.commit:
      try:
        CLIENTS["iam"].put_role_policy(RoleName=role_name, PolicyName=policy_name, PolicyDocument=policy_document)
      except:
        fail("Exception putting policy: {} onto role: {}".format(policy_name, role_name), sys.exc_info())

def conditionally_create_kms_key(role_name, service_type):
  """
  Create KMS Master Key for encryption/decryption of sensitive values in cf templates and latebind configs
  Args:
      role_name: name of the role that kms key is being created for; it will be given decrypt privileges.
      service_type: service registry service type: 'aws_ec2', 'aws_ecs', 'aws_ecs_http', 'aws_fixture',
                                                   'aws_lambda', or 'http_service'
  """
  if service_type not in KMS_SERVICE_TYPES:
    print_if_verbose("not eligible for kms; service_type: {} is not valid for kms".format(service_type))
    return

  # Converting all periods to underscores because they are invalid in KMS alias names
  key_alias = role_name.replace('.', '_')

  try:
    kms_key = CLIENTS["kms"].describe_key(KeyId='alias/{}'.format(key_alias))
  except ClientError as error:
    if error.response['Error']['Code'] == 'NotFoundException':
      kms_key = None
    else:
      fail("Exception describing KMS key: {} {}".format(role_name, error))

  account_id = CONTEXT.account_id
  if service_type == "aws_fixture":
    kms_key_policy = {
      "Version": "2012-10-17",
      "Statement": [
        {
          "Sid": "Enable IAM User Permissions",
          "Effect": "Allow",
          "Principal": {
            "AWS": "arn:aws:iam::{}:root".format(account_id)
          },
          "Action": "kms:*",
          "Resource": "*"
        }
      ]
    }
  else:
    role_principal = [ "arn:aws:iam::{}:role/{}".format(account_id, role_name) ]
    if '.' not in role_name:
      subservice_principals = "arn:aws:iam::{}:role/{}.*".format(account_id, role_name)
      role_principal.append(subservice_principals)
    kms_key_policy = {
      "Version": "2012-10-17",
      "Statement": [
        {
          "Sid": "Enable IAM User Permissions",
          "Effect": "Allow",
          "Principal": {
            "AWS": "arn:aws:iam::{}:root".format(account_id)
          },
          "Action": "kms:*",
          "Resource": "*"
        },
        {
          "Sid": "Allow Service Role Decrypt Privileges",
          "Effect": "Allow",
          "Principal": "*",
          "Action": "kms:Decrypt",
          "Resource": "*",
          "Condition": {
                "ForAnyValue:StringLike": {
                    "aws:PrincipalArn": role_principal
                }
            }
        },
        {
          "Sid": "Allow use of the key for default autoscaling group service role",
          "Effect": "Allow",
          "Principal": { "AWS": "arn:aws:iam::{}:role/aws-service-role/autoscaling.amazonaws.com/AWSServiceRoleForAutoScaling".format(account_id) },
          "Action": [
            "kms:Encrypt",
            "kms:Decrypt",
            "kms:ReEncrypt*",
            "kms:GenerateDataKey*",
            "kms:DescribeKey"
          ],
          "Resource": "*"
        },
        {
          "Sid": "Allow attachment of persistent resourcesfor default autoscaling group service role",
          "Effect": "Allow",
          "Principal": { "AWS": "arn:aws:iam::{}:role/aws-service-role/autoscaling.amazonaws.com/AWSServiceRoleForAutoScaling".format(account_id) },
          "Action": [
            "kms:CreateGrant"
          ],
          "Resource": "*",
          "Condition": {
            "Bool": {
              "kms:GrantIsForAWSResource": True
            }
          }
        }
      ]
    }

  if not kms_key:
    print("Create KMS key: {}".format(key_alias))
    if CONTEXT.commit:
      # Create KMS Master Key. Due to AWS eventual consistency a newly created IAM role may not be
      # immediately visible to KMS. Retrying up to 5 times (25 seconds) to account for this behavior.
      create_key_failures = 0
      while create_key_failures <= 5:
        try:
          kms_key = CLIENTS["kms"].create_key(
            Policy=json.dumps(kms_key_policy, indent=4),
            Description='Master Key for {}'.format(role_name)
          )
          break
        except ClientError as error:
          if error.response['Error']['Code'] == 'MalformedPolicyDocumentException':
            if create_key_failures == 5:
              fail("Exception creating kms key: {} {}".format(role_name, error))
            else:
              create_key_failures += 1
              time.sleep(5)
          else:
            fail("Exception creating kms key: {} {}".format(role_name, error))

      # Assign key an alias. This is used for all future references to it (rather than the key ARN)
      try:
        CLIENTS["kms"].create_alias(
          AliasName='alias/{}'.format(key_alias),
          TargetKeyId=kms_key['KeyMetadata']['KeyId']
        )
      except ClientError as error:
        fail("Exception creating alias for kms key: {} {}".format(role_name, error))
  else:
    print_if_verbose("KMS key already exists: {}".format(key_alias))

  # Enable KMS key rotation
  if CONTEXT.commit:
    try:
      CLIENTS["kms"].enable_key_rotation(KeyId=kms_key["KeyMetadata"]["KeyId"])
    except ClientError as error:
      fail("Error in enabling key rotation: {} {}".format(role_name, error))


def create_newrelic_alerts():
  """
   Create Newrelic Alerts for each entry in the service registry application_services
   Note: Import is inside this function rather than top of the file so that we do not import when running unit tests.
         (would otherwise require an ef_site_config.yml in the directory where tests are being run)
  """
  from newrelic_executor import NewRelicAlerts
  print("> Creating NewRelic Alerts")
  NewRelicAlerts(CONTEXT, CLIENTS).run()

def main():
  global CONTEXT, CLIENTS, AWS_RESOLVER

  CONTEXT = handle_args_and_set_context(sys.argv[1:])
  if not CONTEXT.devel and CONTEXT.whereami != 'jenkins':
    try:
      pull_repo()
    except RuntimeError as error:
      fail("Error checking or pulling repo", error)
  else:
    print("Not refreshing repo because --devel was set or running on Jenkins")

  # sign on to AWS and create clients and get account ID
  try:
    # If running in EC2, always use instance credentials. One day we'll have "lambda" in there too, so use "in" w/ list
    if CONTEXT.whereami == "ec2":
      CLIENTS = create_aws_clients(EFConfig.DEFAULT_REGION, None, "cloudfront", "ec2", "iam", "kms")
      CONTEXT.account_id = str(json.loads(http_get_metadata('iam/info'))["InstanceProfileArn"].split(":")[4])
    else:
      # Otherwise, we use local user creds based on the account alias
      CLIENTS = create_aws_clients(EFConfig.DEFAULT_REGION, CONTEXT.account_alias, "cloudfront", "ec2", "iam", "kms", "sts")
      CONTEXT.account_id = get_account_id(CLIENTS["sts"])
  except RuntimeError:
    fail("Exception creating AWS clients in region {} with profile {}".format(
      EFConfig.DEFAULT_REGION, CONTEXT.account_alias))
  # Instantiate an AWSResolver to lookup AWS resources
  AWS_RESOLVER = EFAwsResolver(CLIENTS)

  # Show where we're working
  if not CONTEXT.commit:
    print("=== DRY RUN ===\nUse --commit to create roles and security groups\n=== DRY RUN ===")
  print("env: {}".format(CONTEXT.env))
  print("env_full: {}".format(CONTEXT.env_full))
  print("env_short: {}".format(CONTEXT.env_short))
  print("aws account profile: {}".format(CONTEXT.account_alias))
  print("aws account number: {}".format(CONTEXT.account_id))

  # If we're only ef-generate for a specific service, print it out here
  if CONTEXT.only_for:
    print("running ef-generate only for entry: {}".format(CONTEXT.only_for))

  # Step through all services in the service registry
  for CONTEXT.service in CONTEXT.service_registry.iter_services():
    service_name = CONTEXT.service[0]
    target_name = "{}-{}".format(CONTEXT.env, service_name)
    sr_entry = CONTEXT.service[1]
    service_type = sr_entry['type']
    print_if_verbose("service: {} in env: {}".format(service_name, CONTEXT.env))

    # Are we targeting a specific entry in the registry?
    if CONTEXT.only_for and CONTEXT.only_for != service_name:
      continue

    # Is this service_type handled by this tool?
    if service_type not in SUPPORTED_SERVICE_TYPES:
      print_if_verbose("unsupported service type: {}".format(service_type))
      continue
    # Is the env valid for this service?
    if CONTEXT.env_full not in CONTEXT.service_registry.valid_envs(service_name):
      print_if_verbose("env: {} not valid for service {}".format(CONTEXT.env_full, service_name))
      continue
    # Is the service_type allowed in 'global'?
    if CONTEXT.env == "global" and service_type not in GLOBAL_SERVICE_TYPES:
      print_if_verbose("env: {} not valid for service type {}".format(CONTEXT.env, service_type))
      continue

    # 1. CONDITIONALLY MAKE ROLE AND/OR INSTANCE PROFILE FOR THE SERVICE
    # If service gets a role, create with either a custom or default AssumeRole policy document
    conditionally_create_role(target_name, sr_entry, tags=sr_entry.get('tags'))
    # Instance profiles and security groups are not allowed in the global scope
    if CONTEXT.env != "global":
      conditionally_create_profile(target_name, service_type)

      # 2. SECURITY GROUP(S) FOR THE SERVICE : only some types of services get security groups
      conditionally_create_security_groups(CONTEXT.env, service_name, service_type)

    # 3. KMS KEY FOR THE SERVICE : only some types of services get kms keys
    conditionally_create_kms_key(target_name, service_type)

    # 4. ATTACH AWS MANAGED POLICIES TO ROLE
    conditionally_attach_aws_managed_policies(target_name, sr_entry)

    # 5. ATTACH CUSTOMER MANAGED POLICIES TO ROLE
    conditionally_attach_customer_managed_policies(target_name, sr_entry)

    # 6. INLINE SERVICE'S POLICIES INTO ROLE
    # only eligible service types with "policies" sections in the service registry get policies
    conditionally_inline_policies(target_name, sr_entry)

    # Break from the loop if we've finished working on the targeted entry
    if CONTEXT.only_for:
      break

  # Create newrelic alerts for all "application_services" in the service registry
  if "newrelic" in EFConfig.PLUGINS:
    create_newrelic_alerts()

  print("Exit: success")


if __name__ == "__main__":
  main()



"""
Instance initialization
This is installed on EC2 instances and local dev to load configuration values at startup.
Lambda initialization is not available yet.

For config object paths in AWS, see:
https://ellation.atlassian.net/wiki/display/DEVOPS/VRV+Name+Patterns+and+Paths


Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function
from grp import getgrnam
from syslog import closelog, openlog, syslog
from os import chmod, chown, makedirs
from os.path import dirname, normpath
from pwd import getpwnam
from socket import gethostname
import sys

import boto3
import boto3.utils
import botocore.exceptions

from ef_config import EFConfig
from ef_instanceinit_config_reader import EFInstanceinitConfigReader
from ef_utils import http_get_instance_role, http_get_metadata, whereami
from ef_conf_utils import get_account_alias
from ef_template_resolver import EFTemplateResolver

# constants
LOG_IDENT = "ef-instanceinit"
VIRTUALBOX_CONFIG_ROOT = "/vagrant/configs"

# globals
RESOURCES = {}  # boto resources (easier to use for some things)
WHERE = None


def log_info(message):
  """
  Log a message to log_info and console, and return
  Args:
    message:
  """
  print(message)
  syslog(message)


def critical(message):
  """
  Log critical error to log_info and console and exit with error status
  """
  log_info(message)
  closelog()
  sys.exit(1)


def get_user_group(dest):
  """
  Given a dictionary object representing the dest JSON in the late bind config's parameter file, return two
  values, the user and group
  Args:
      dest: dict object from the late bind config's parameters file e.g. dest["user_group"] = "Bob:devops"

  Returns:
      user: user that the late bind config belongs to
      group: group that the late bind config belongs to

  """
  return dest["user_group"].split(":")


def merge_files(service, skip_on_user_group_error=False):
  """
  Given a prefix, find all templates below; merge with parameters; write to "dest"
  Args:
    service: "<service>", "all", or "ssh"
    skip_on_user_group_error: True or False

  For S3, full path becomes:
    s3://ellation-cx-global-configs/<service>/templates/<filename>
    s3://ellation-cx-global-configs/<service>/parameters/<filename>.parameters.<yaml|yml|json>
  For filesystem, full path becomes:
    /vagrant/configs/<service>/templates/<filename>
    /vagrant/configs/<service>/parameters/<filename>.parameters.<yaml|yml|json>
  """
  if WHERE == "ec2":
    config_reader = EFInstanceinitConfigReader("s3", service, log_info, RESOURCES["s3"])
    resolver = EFTemplateResolver()
  elif WHERE == "virtualbox-kvm":
    config_path = "{}/{}".format(VIRTUALBOX_CONFIG_ROOT, service)
    config_reader = EFInstanceinitConfigReader("file", config_path, log_info)
    environment = EFConfig.VAGRANT_ENV
    resolver = EFTemplateResolver(env=environment, profile=get_account_alias(environment),
                                  region=EFConfig.DEFAULT_REGION, service=service)

  while config_reader.next():
    log_info("checking: {}".format(config_reader.current_key))

    # if 'dest' for the current object contains an 'environments' list, check it
    dest = config_reader.dest
    if "environments" in dest:
      if not resolver.resolved["ENV_SHORT"] in dest["environments"]:
        log_info("Environment: {} not enabled for {}".format(
          resolver.resolved["ENV_SHORT"], config_reader.current_key)
        )
        continue

    # If 'dest' for the current object contains a user_group that hasn't been created in the environment yet and the
    # flag is set to True to skip, log the error and move onto the next config file without blowing up.
    if skip_on_user_group_error:
      user, group = get_user_group(dest)
      try:
        getpwnam(user).pw_uid
      except KeyError:
        log_info("File specifies user {} that doesn't exist in environment. Skipping config file.".format(user))
        continue
      try:
        getgrnam(group).gr_gid
      except KeyError:
        log_info("File specifies group {} that doesn't exist in environment. Skipping config file.".format(group))
        continue

    # Process the template_body - apply context + parameters
    log_info("Resolving template")
    resolver.load(config_reader.template, config_reader.parameters)
    rendered_body = resolver.render()
    if not resolver.resolved_ok():
      critical("Couldn't resolve all symbols; template has leftover {{ or }}: {}".format(resolver.unresolved_symbols()))

    # Write the rendered file
    dir_path = normpath(dirname(dest["path"]))
    # Resolved OK. try to write the template
    log_info("make directories: {} {}".format(dir_path, dest["dir_perm"]))
    try:
      makedirs(dir_path, int(dest["dir_perm"], 8))
    except OSError as error:
      if error.errno != 17:
        critical("Error making directories {}".format(repr(error)))
    log_info("open: " + dest["path"] + ",w+")
    try:
      outfile = open(dest["path"], 'w+')
      log_info("write")
      outfile.write(rendered_body)
      log_info("close")
      outfile.close()
      log_info("chmod file to: " + dest["file_perm"])
      chmod(dest["path"], int(dest["file_perm"], 8))
      user, group = get_user_group(dest)
      uid = getpwnam(user).pw_uid
      gid = getgrnam(group).gr_gid
      log_info("chown last directory in path to: " + dest["user_group"])
      chown(dir_path, uid, gid)
      log_info("chown file to: " + dest["user_group"])
      chown(dest["path"], uid, gid)
    except Exception as error:
      critical("Error writing file: " + dest["path"] + ": " + repr(error))


def main():
  global RESOURCES, WHERE

  openlog(LOG_IDENT)

  WHERE = whereami()
  if WHERE == "unknown":
    critical("Cannot determine whether operating context is ec2 or local. Exiting.")
  elif WHERE == "local":
    critical("local mode not supported: must run under virtualbox-kvm or in ec2")
  elif WHERE == 'jenkins':
    critical("will not run inside a Jenkins environment")
  elif WHERE == "ec2":
    # get info needed for logging
    try:
      instance_id = http_get_metadata('instance-id')
      log_info("Instance id: " + instance_id)
    except IOError as error:
      critical("Error retrieving instance-id metadata: " + repr(error))

  log_info("startup")
  log_info("boto3: " + str(boto3.utils.sys.version_info))
  log_info("boto3: " + str(boto3.utils.sys.version))
  # Tailor to operating mode
  if WHERE == "virtualbox-kvm":
    service = gethostname().split(".", 1)[0]
  elif WHERE == "ec2":
    log_info("EC2: setting up S3 client")
    try:
      session = boto3.Session()
      # S3 iteration is easier using the S3 Resource
      RESOURCES["s3"] = session.resource("s3")
    except (botocore.exceptions.BotoCoreError, IOError) as e:
      critical("Error setting up S3 resource " + repr(e))
    service = http_get_instance_role()

  log_info("platform: {} service: {}".format(WHERE, service))

  merge_files("all")
  merge_files("ssh", skip_on_user_group_error=True)
  merge_files(service)

  log_info("exit: success")
  closelog()


if __name__ == "__main__":
  main()



# noinspection PyClassHasNoInit

"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ef_service_registry import EFServiceRegistry
from ef_utils import whereami
from ef_conf_utils import env_valid
from ef_conf_utils import get_account_alias
from ef_conf_utils import get_env_short
from ef_conf_utils import global_env_valid

class EFContext(object):
  """
  Class holds environment/account-related context for tools and resolvers
  and helps assure consistency in, for example, trimming "env" to "env_short"
  """

  def __init__(self):
    # service environment
    self._account_alias = None
    self._env = None # the name used in AWS -- e.g. prod, proto<n>, mgmt
    self._env_short = None # the generic name -- e.g. prod, proto, mgmt
    self._env_full = None # name found in SR -- e.g. prod, proto, mgmt.<account_alias>
    self._service = None
    # Service registry object
    self._service_registry = None
    # tool context
    self._account_id = None
    self._aws_clients = None
    self._commit = None
    self._devel = None
    self._only_for = None
    self._verbose = None
    self._whereami = whereami()


  @property
  def account_alias(self):
    """The account alias in use"""
    return self._account_alias

  @property
  def env(self):
    """Full name of the environment, e.g. 'prod' or 'proto3'"""
    return self._env

  @env.setter
  def env(self, value):
    """
    Sets context.env, context.env_short, and context.account_alias if env is valid
    For envs of the form "global.<account>" and "mgmt.<account_alias>",
    env is captured as "global" or "mgmt" and account_alias is parsed
    out of the full env rather than looked up
    Args:
      value: the fully-qualified env value
    Raises:
      ValueError if env is not valid
    """
    env_valid(value)
    self._env_full = value
    if value.find(".") == -1:
      # plain environment, e.g. prod, staging, proto<n>
      self._env = value
      self._account_alias = get_account_alias(value)
    else:
      # "<env>.<account_alias>" form, e.g. global.ellationeng or mgmt.ellationeng
      self._env, self._account_alias = value.split(".")
      # since we extracted an env, must reconfirm that it's legit
      global_env_valid(self._env)
    self._env_short = get_env_short(value)

  @property
  def env_short(self):
    """Short (generic) name of the environment, e.g. 'prod' or 'proto' or 'mgmt'"""
    return self._env_short

  @property
  def env_full(self):
    """Name of the environment as expected in the service registry, e.g. 'prod' or 'proto' or 'mgmt.ellationeng'"""
    return self._env_full

  @property
  def service(self):
    """A single service's object from service registry"""
    return self._service

  @service.setter
  def service(self, value):
    self._service = value

  @property
  def service_registry(self):
    """Service registry object"""
    return self._service_registry

  @service_registry.setter
  def service_registry(self, sr):
    """
    Sets service registry object in context, doesn't check it
    Args:
      sr: EFServiceRegistry object
    """
    if type(sr) is not EFServiceRegistry:
      raise TypeError("sr value must be type 'EFServiceRegistry'")
    self._service_registry = sr

  @property
  def account_id(self):
    """
    Retrieves the current account id

    Returns:
      account id (string)
    """
    return self._account_id

  @account_id.setter
  def account_id(self, value):
    """
    Sets the current account id

    Args:
      value: current account id (string)

    Returns:
      None
    """
    if type(value) is not str:
      raise TypeError("commit value must be string")
    self._account_id = value

  def aws_client(self, client_id=None):
    """
    Get AWS client if it exists (must have been formerly stored with set_aws_clients)
    If client_id is not provided, returns the dictionary of all clients
    Args:
      client_id: label for the client, e.g. 'ec2'; omit to get a dictionary of all clients
    Returns:
      aws client if found, or None if not
    """
    if client_id is None:
      return self._aws_clients
    elif self._aws_clients is not None and client_id in self._aws_clients:
      return self._aws_clients[client_id]
    else:
      return None

  def set_aws_clients(self, clients):
    """
    Stash a dictionary of AWS clients in the context object
    Args:
      clients: dictionary of clients
    """
    if type(clients) is not dict:
      raise TypeError("clients must be a dict")
    self._aws_clients = clients

  @property
  def commit(self):
    """True if the tool should actually execute changes"""
    return self._commit

  @commit.setter
  def commit(self, value):
    if type(value) is not bool:
      raise TypeError("commit value must be bool")
    self._commit = value

  @property
  def devel(self):
    """True if the tool should allow devel exceptions"""
    return self._devel

  @devel.setter
  def devel(self, value):
    if type(value) is not bool:
      raise TypeError("devel value must be bool")
    self._devel = value

  @property
  def only_for(self):
    """Specific service registry entry targeted as parameter passed to tool"""
    return self._only_for

  @only_for.setter
  def only_for(self, value):
    self._only_for = value

  @property
  def verbose(self):
    """True if the tool should print extra info"""
    return self._verbose

  @verbose.setter
  def verbose(self, value):
    if type(value) is not bool:
      raise TypeError("verbose value must be bool")
    self._verbose = value

  @property
  def whereami(self):
    """Hosted ec2? lambda? local vm?"""
    return self._whereami



#!/usr/bin/env python

"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

from ef_template_resolver import EFTemplateResolver
from ef_conf_utils import get_account_alias

PARAMS = """{
  "params":{
    "default":{
      "address": "default address",
      "phone": "default phone",
      "ACC": "ACC",
      "OUNT": "OUNT",
      "/_-.": "slashunderscoredashdot",
      "illegal": "cannot have a comma",
      ".": "dot",
      "my-thing": "my-thing",
      "multi-line-thing": "line1\\nline2\\nline3\\nline4",
      "multi-line-thing-2": [
        "first thing",
        "second thing"
      ]
    },
    "proto":{
      "blah": "this text should only print if env is protoN",
      "name": "proto name"
    },
    "proto3":{
      "name": "proto3 name",
      "address": "proto3 address",
      "ENV": "this should never print"
    },
    "staging": {
      "name": "staging name"
    }
  }
}
"""

LOCAL = True

TEST_STRING = "name: {{name}} address:{{address}} phone:{{phone}}?\n \
account:{{ACCOUNT}},\n \
nested account:{{{{ACC}}{{OUNT}}}}, env:{{ENV}},\n \
fully-qualified environment:{{ENV_FULL}},\n \
region:{{REGION}},\n \
special characters:{{/_-.}},\n \
dot:{{.}},\n\
general 'proto' token only if env is protoN: {{blah}}\n \
and this {{DOESNOTRESOLVE}} and {{neitherdoesthis}}\n\
and {{this is not a symbol}} and {{thisisalsonotasymbol?!}}\n\
hyphenated symbol name: {{my-thing}}\n\
aws security group lookup: {{aws:ec2:security-group/security-group-id,proto2-core-elb}}\n\
aws lookup AZs of staging env via vpc+subnets: {{aws:ec2:vpc/availabilityzones,vpc-staging}}\n\
aws lookup AZs of proto3 env via vpc+subnets: {{aws:ec2:vpc/availabilityzones,vpc-proto3}}\n\
aws lookup subnets of staging env via vpc: {{aws:ec2:vpc/subnets,vpc-staging}}\n\
aws lookup subnets of proto3 env via vpc: {{aws:ec2:vpc/subnets,vpc-proto3}}\n\
aws lookup subnets of the current env '{{ENV}}' via vpc: {{aws:ec2:vpc/subnets,vpc-{{ENV}}}}\n\
multi-line:\n\
{{multi-line-thing}}\n\
\"{{multi-line-thing-2}}\"\n\
Subnets: [ \"{{aws:ec2:vpc/subnets,vpc-staging}}\" ]\n\
Single-subnet lookup, subnet-proto3-a: {{aws:ec2:subnet/subnet-id,subnet-{{ENV}}-a}}\n\
WAF Rule ID lookup: {{aws:waf:rule-id,global-OfficeCidr}}\n\
Public Hosted Zone lookup: {{aws:route53:public-hosted-zone-id,cx-proto3.com.}}\n\
Private Hosted Zone lookup: {{aws:route53:private-hosted-zone-id,cx-proto3.com.}}\n\
Route table lookup: {{aws:ec2:route-table/main-route-table-id,vpc-proto0}}\n\
Cloudfront distribution Domain Name: {{aws:cloudfront:domain-name,api.cx-proto3.com}}\n\
Cloudfront Origin Access Identity ID lookup: \
{{aws:cloudfront:origin-access-identity/oai-id,static.cx-proto3.com}}\n\
Cloudfront Origin Access Identity Canonical User ID lookup: \
{{aws:cloudfront:origin-access-identity/oai-canonical-user-id,static.cx-proto3.com}}\n\
VPC CIDR block: {{aws:ec2:vpc/cidrblock,vpc-staging}}\n\
WAF Web ACL ID: {{aws:waf:web-acl-id,staging-StaticAcl}}\n\
SSL Certificate ARN us-west-2/cx-proto3.com: {{aws:acm:certificate-arn,us-west-2/cx-proto3.com}}\n\
Elastic network interface (ENI) eni-proto0-dnsproxy-1a: {{aws:ec2:eni/eni-id,eni-proto0-dnsproxy-1a}}\n\
Elastic IP Allocation ID: {{aws:ec2:elasticip/elasticip-id,ElasticIpMgmtCingest1}}\n\
Elastic IP IP Address: {{aws:ec2:elasticip/elasticip-ipaddress,ElasticIpMgmtCingest1}}\n\
EFConfig resolver, accountaliasofenv,prod: {{efconfig:accountaliasofenv,staging}}\n\
AMI lookup: {{version:ami-id,proto0/test-instance}}\n\
Latest AMI for test-instance: {{version:ami-id,proto0/test-instance}}\n\
Custom Data: \"{{efconfig:customdata,office_ips}}\"\
"

GLOBAL_ENV_TEST_STRING = "fully-qualified environment:{{ENV_FULL}}\n"

# Test with proto0
if LOCAL:
  resolver = EFTemplateResolver(profile=get_account_alias("proto0"), env="proto0", region="us-west-2",
                                service="mine", verbose=True)
else:
  resolver = EFTemplateResolver(verbose=True)

resolver.load(TEST_STRING, PARAMS)
resolver.render()

print(resolver.template)
print("unresolved symbol count: "+str(len(resolver.unresolved_symbols())))
print("unresolved symbols: "+repr(resolver.unresolved_symbols()))
print("all template symbols: "+repr(resolver.symbols))
print("all EFTemplateResolver symbols: "+repr(resolver.resolved))


# Demo with the global env 'mgmt.ellationeng' (local only)
if LOCAL:
  resolver = EFTemplateResolver(profile=get_account_alias("mgmt.ellationeng"), env="mgmt.ellationeng", region="us-west-2",
                                service="mine", verbose=True)

  resolver.load(GLOBAL_ENV_TEST_STRING, PARAMS)
  resolver.render()

  print("\nDemo 'mgmt.ellationeng' resolution")
  print(resolver.template)
  print("unresolved symbol count: "+str(len(resolver.unresolved_symbols())))
  print("unresolved symbols: "+repr(resolver.unresolved_symbols()))
  print("all template symbols: "+repr(resolver.symbols))
  print("all EFTemplateResolver symbols: "+repr(resolver.resolved))



# noinspection PyClassHasNoInit

"""
Copyright 2016-2017 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ef_site_config import EFSiteConfig


class EFConfig(object):
  """
  Installation-specific and global settings shared by all EF tools
  Don't change anything here.
  All supported site-specific customizations are found in ef_site_config.py
  """

  _ef_site_config = EFSiteConfig().load()

  # Initialize config constants
  ALLOW_EF_VERSION_SKIP_PRECHECK = _ef_site_config["ALLOW_EF_VERSION_SKIP_PRECHECK"]
  if "CUSTOM_DATA" in _ef_site_config:
    CUSTOM_DATA = _ef_site_config["CUSTOM_DATA"]
  DEFAULT_REGION = _ef_site_config["DEFAULT_REGION"]
  EF_CF_POLL_PERIOD = _ef_site_config["EF_CF_POLL_PERIOD"]
  EF_REPO_BRANCH = _ef_site_config["EF_REPO_BRANCH"]
  ENV_ACCOUNT_MAP = _ef_site_config["ENV_ACCOUNT_MAP"]
  EPHEMERAL_ENVS = _ef_site_config["EPHEMERAL_ENVS"]
  S3_CONFIG_BUCKET = _ef_site_config["S3_CONFIG_BUCKET"]
  S3_VERSION_BUCKET = _ef_site_config["S3_VERSION_BUCKET"]
  SERVICE_GROUPS = set(_ef_site_config["SERVICE_GROUPS"])
  SPECIAL_VERSION_ENVS = _ef_site_config["SPECIAL_VERSION_ENVS"]
  STACK_TERMINATION_PROTECTED_ENVS = _ef_site_config["STACK_TERMINATION_PROTECTED_ENVS"]
  VAGRANT_ENV = _ef_site_config["VAGRANT_ENV"]
  PLUGINS = _ef_site_config.get("PLUGINS", {})

  # Default service registry file name
  DEFAULT_SERVICE_REGISTRY_FILE = "service_registry.json"
  PARAMETER_FILE_SUFFIXES = [".parameters.yaml", ".parameters.yml", ".parameters.json"]
  POLICY_TEMPLATE_PATH_SUFFIX = "/policy_templates/"
  # the service group 'fixtures' always exists
  SERVICE_GROUPS.add("fixtures")

  # Convenient list of all mapped accounts
  ACCOUNT_ALIAS_LIST = set(ENV_ACCOUNT_MAP.values())

  # These environments are for account-wide resources; they have a ".<ACCOUNT_ALIAS>" suffix
  ACCOUNT_SCOPED_ENVS = ["global", "mgmt"]

  # Protected environments are unique and non-ephemeral
  PROTECTED_ENVS = ["prod"] + ACCOUNT_SCOPED_ENVS

  # Convenient list of all possible valid environments
  ENV_LIST = []
  VALID_ENV_REGEX = ""
  for env in ENV_ACCOUNT_MAP.keys():
    if env not in PROTECTED_ENVS and env in EPHEMERAL_ENVS:
      ENV_LIST.extend((lambda env=env, EPHEMERAL_ENVS=EPHEMERAL_ENVS: [env + str(x) for x in range(EPHEMERAL_ENVS[env])])())
      VALID_ENV_REGEX += "{}[0-{}]|".format(env, EPHEMERAL_ENVS[env] - 1)
    else:
      ENV_LIST.append(env)
      VALID_ENV_REGEX += "{}|".format(env)

  ENV_LIST.extend("global." + x for x in ACCOUNT_ALIAS_LIST)
  ENV_LIST.extend("mgmt." + x for x in ACCOUNT_ALIAS_LIST)
  ENV_LIST = sorted(ENV_LIST)
  VALID_ENV_REGEX += "global|mgmt"

  # Version system
  # suffix used for naming deployable service AMIs
  AMI_SUFFIX = "-release"
  # content-encoding for S3 version registry
  S3_VERSION_CONTENT_ENCODING = "utf-8"
  # Metdata key on a version object to indicate who modified it
  S3_VERSION_BUILDNUMBER_KEY = "ef-buildnumber"
  # Metdata key on a version object to indicate who modified it
  S3_VERSION_COMMITHASH_KEY = "ef-commithash"
  # Metdata key on a version object to indicate who modified it
  S3_VERSION_LOCATION_KEY = "ef-location"
  # Metdata key on a version object to indicate who modified it
  S3_VERSION_MODIFIEDBY_KEY = "ef-modifiedby"
  # Metadata key on a version object to indicate its status
  S3_VERSION_STATUS_KEY = "ef-version-status"
  # Metadata version status values
  S3_VERSION_STATUS_STABLE = "stable"
  S3_VERSION_STATUS_UNDEFINED = "undefined"
  VERSION_KEYS = {
      "ami-id": {
          "allow_latest": True,
          "allowed_types": ["aws_ec2", "http_service"]
      },
      "image-tag": {
        "allow_latest": True,
        "allowed_types": ["aws_ecs", "aws_ecs_http"]
      },
      "commit-hash": {
        "allow_latest": True,
        "allowed_types": ["aws_lambda"]
      },
      "config": {},
      "dist-hash": {
          "allowed_types": ["dist_static"]
      },
      "function-version": {
        # TODO: remove once all templates have moved to version-number
        "allow_latest": True,
        "allowed_types": ["aws_lambda"]
      },
      "version-number": {
        "allow_latest": True,
        "allowed_types": ["aws_lambda"]
      },
  }
  # Some envs' version entries can be set via these special values, meaning 'use the value found there'
  SPECIAL_VERSIONS = ["=latest", "=prod", "=staging"]



"""
Copyright 2016-2018 Ellation, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

from ef_config import EFConfig


class EFConfigResolver(object):
  """
  Resolves values from the tool configuration in the EFConfig class (ef_config.py)

  In a template:
    {{efconfig:thing,lookup}}
    {{efconfig:accountaliasofenv,prod}} <-- gets the account alias of the account that hosts the 'prod' env
  """

  def accountaliasofenv(self, lookup, default=None):
    """
    Args:
      lookup: ENV_SHORT name of an env, such as: 'prod' or 'proto'
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The account alias of the account that hosts the env named in lookupor default/None if no match found
    """
    if lookup in EFConfig.ENV_ACCOUNT_MAP:
      return EFConfig.ENV_ACCOUNT_MAP[lookup]
    else:
      return None

  def customdata(self, lookup, default=None):
    """
    Args:
      lookup: the custom data file
      default: the optional value to return if lookup failed; returns None if not set
    Returns:
      The custom data returned from the file 'lookup' or default/None if no match found
    """
    try:
      if lookup in EFConfig.CUSTOM_DATA:
        return EFConfig.CUSTOM_DATA[lookup]
      else:
        return default
    except AttributeError:
      return default

  def lookup(self, token):
    try:
      kv = token.split(",")
    except ValueError:
      return None
    if kv[0] == "accountaliasofenv":
      return self.accountaliasofenv(*kv[1:])
    if kv[0] == "customdata":
      return self.customdata(*kv[1:])
    else:
      return None

  def __init__(self):
    pass



"""
Configurable utility functions for ef. Configurable via EFConfig
"""
import re
import subprocess
from os.path import exists

from botocore.exceptions import ClientError

from ef_config import EFConfig

def get_template_parameters_file(template_full_path):
    """
    Checks for existance of parameters file against supported suffixes and returns parameters file path if found
    Args:
      template_full_path: full filepath for template file
    Returns:
      filename of parameters file if it exists
    """
    for suffix in EFConfig.PARAMETER_FILE_SUFFIXES:
      parameters_file = template_full_path.replace("/templates", "/parameters") + suffix
      if exists(parameters_file):
        return parameters_file
    return None

def global_env_valid(env):
  """
  Given an env, determine if it's a valid "global" or "mgmt" env as listed in EFConfig
  Args:
    env: the env to check
  Returns:
    True if the env is a valid global env in EFConfig
  Raises:
    ValueError with message if the env is not valid
  """
  if env not in EFConfig.ACCOUNT_SCOPED_ENVS:
    raise ValueError("Invalid global env: {}; global envs are: {}".format(env, EFConfig.ACCOUNT_SCOPED_ENVS))
  return True

def env_valid(env):
  """
  Given an env, determine if it's valid
  Args:
    env: the env to check
  Returns:
    True if the env is valid
  Raises:
    ValueError with message if the env is not valid
  """
  if env not in EFConfig.ENV_LIST:
    raise ValueError("unknown env: {}; env must be one of: ".format(env) + ", ".join(EFConfig.ENV_LIST))
  return True

def pull_repo():
  """
  Pulls latest version of EF_REPO_BRANCH
  Raises:
    RuntimeError with message if not on the correct branch
  """
  try:
    current_branch = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).rstrip()
  except subprocess.CalledProcessError as error:
    raise RuntimeError("Exception checking current branch: " + repr(error))
  if current_branch != EFConfig.EF_REPO_BRANCH:
    raise RuntimeError("Must be on branch: " + EFConfig.EF_REPO_BRANCH + ". Current branch is: " + current_branch)
  try:
    subprocess.check_call(["git", "pull", "-q", "origin", EFConfig.EF_REPO_BRANCH])
  except subprocess.CalledProcessError as error:
    raise RuntimeError("Exception running 'git pull': " + repr(error))

def get_account_alias(env):
  """
  Given an env, return <account_alias> if env is valid
  Args:
    env: an environment, such as "prod", "staging", "proto<N>", "mgmt.<account_alias>"
  Returns:
    the alias of the AWS account that holds the env
  Raises:
    ValueError if env is misformatted or doesn't name a known environment
  """
  env_valid(env)
  # Env is a global env of the form "env.<account_alias>" (e.g. "mgmt.<account_alias>")
  if env.find(".") > -1:
    base, ext = env.split(".")
    return ext
  # Ordinary env, possibly a proto env ending with a digit that is stripped to look up the alias
  else:
    env_short = env.strip(".0123456789")
    if env_short not in EFConfig.ENV_ACCOUNT_MAP:
      raise ValueError("generic env: {} has no entry in ENV_ACCOUNT_MAP of ef_site_config.py".format(env_short))
    return EFConfig.ENV_ACCOUNT_MAP[env_short]

def get_template_parameters_s3(template_key, s3_resource):
  """
  Checks for existance of parameters object in S3 against supported suffixes and returns parameters file key if found
  Args:
    template_key: S3 key for template file. omit bucket.
    s3_resource: a boto3 s3 resource
  Returns:
    filename of parameters file if it exists
  """
  for suffix in EFConfig.PARAMETER_FILE_SUFFIXES:
    parameters_key = template_key.replace("/templates", "/parameters") + suffix
    try:
      obj = s3_resource.Object(EFConfig.S3_CONFIG_BUCKET, parameters_key)
      obj.get()
      return parameters_key
    except ClientError:
      continue
  return None

def get_env_short(env):
  """
  Given an env, return <env_short> if env is valid
  Args:
    env: an environment, such as "prod", "staging", "proto<N>", "mgmt.<account_alias>"
  Returns:
    the shortname of the env, such as "prod", "staging", "proto", "mgmt"
  Raises:
    ValueError if env is misformatted or doesn't name a known environment
  """
  env_valid(env)
  if env.find(".") > -1:
    env_short, ext = env.split(".")
  else:
    env_short = env.strip(".0123456789")
  return env_short
